"""
@file
@brief One class which visits a syntax tree.
"""

import ast
from .onnx_translator import OnnxTranslator


class CodeNodeVisitor(ast.NodeVisitor):

    """
    Defines a visitor which walks though the syntax tree of the code.

    .. exref::
        :title: Get the tree of a simple function

        The following code uses Python syntax but follows a SQL logic.

        .. runpython::
            :showcode:
            :process:
            :store_in_file: fct2onnx1.py

            import ast
            import inspect
            from textwrap import dedent
            from mlprodict.onnx_grammar import CodeNodeVisitor

            def norm2(x, y):
                delta = x - y
                n = delta ** 2
                return n

            code = dedent(inspect.getsource(norm2))
            node = ast.parse(code)
            v = CodeNodeVisitor()
            v.visit(node)
            for r in v.Rows :
                print("{0}{1}: {2}".format("    " * r["indent"], r["type"], r["str"]))
    """

    def __init__(self, translator=None):
        """
        @param      translator      @see cl CodeTranslator

        By default the translator is @see cl OnnxTranslator.
        """
        ast.NodeVisitor.__init__(self)
        self._rows = []
        self._indent = 0
        self._stack = []
        self._translator = OnnxTranslator(
            self) if translator is None else translator

    def push(self, row):
        """
        Pushes an element into a list.
        """
        self._rows.append(row)

    def generic_visit(self, node):
        """
        Overrides ``generic_visit`` to check it is not used.
        """
        raise AttributeError(  # pragma: no cover
            "generic_visit_args should be used.")

    def generic_visit_args(self, node, row):
        """
        Overrides ``generic_visit`` to keep track of the indentation
        and the node parent. The function will add field
        ``row["children"] = visited`` nodes from here.

        @param      node        node which needs to be visited
        @param      row         row (a dictionary)
        @return                 See ``ast.NodeVisitor.generic_visit``
        """
        if hasattr(node, 'lineno'):
            row['lineno'] = node.lineno
        if hasattr(node, 'col_offset'):
            row['col_offset'] = node.col_offset
        self._indent += 1
        last = len(self._rows)
        self._translator.visit(node, row)
        res = ast.NodeVisitor.generic_visit(  # pylint: disable=E1111
            self, node)  # pylint: disable=E1111
        row["children"] = [
            _ for _ in self._rows[
                last:] if _["indent"] == self._indent]
        self._indent -= 1
        self._translator.depart(node, row)
        return res

    def make_msg(self, node):
        """
        Displays line and column information into a string.
        """
        return "line {}, col {}".format(  # pragma: no cover
            getattr(node, 'lineno', '?'), getattr(node, 'col_offset', '?'))

    def visit(self, node):
        """
        Visits a node, a method must exist for every object class.
        """
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, None)
        if visitor is None:
            raise TypeError(  # pragma: no cover
                "Unable to find a method '{}' at {}.".format(
                    method, self.make_msg(node)))
        res = visitor(node)
        # print(method, CodeNodeVisitor.print_node(node))
        return res

    def visit_(self, node):
        """
        If an element is not found...
        """
        raise NotImplementedError(  # pragma: no cover
            "Node '{}' ({}) not recognized at {}\nNode\n{}\n--"
            "Status--\n{}".format(
                node, type(node), self.make_msg(node),
                self.print_node(node), self.print_tree()))

    @staticmethod
    def print_node(node):
        """
        Debugging purpose.
        """
        r = []
        for att in sorted(set(["s", "name", "str", "id", "body", "n",
                               "arg", "targets", "attr", "returns", "ctx",
                               'col_offset', 'lineno',
                               'value'] + list(getattr(node, '_attributes', [])))):
            v = getattr(node, att, None)
            if v is not None or att in getattr(node, '_fields', []):
                r.append("{0}={1}".format(att, v))
        return " ".join(r)

    def print_tree(self):
        """
        Displays the tree of instructions.

        @return     string
        """
        rows = []
        for r in self.Rows:
            rows.append(
                ("{0}{1}: {2}".format(
                    "    " *
                    r["indent"],
                    r["type"],
                    r["str"])))
        return "\n".join(rows)

    @property
    def Rows(self):
        """
        returns a list of dictionaries with all the elements of the code
        """
        return [_ for _ in self._rows if not _.get("remove", False)]

    def export(self, context=None, **kwargs):
        """
        Calls method *export* from the translator class.

        @param      context     known :epkg:`python` needed to run
                                the translated function
        @param      kwargs      whatever the method *export* from
                                the translator class ingests
        @return                 whatever the method *export* from
                                the translator class returns
        """
        return self._translator.export(context=context, **kwargs)

    ###########
    # Methods for python code elements
    ###########

    def visit_Str(self, node):  # pylint: disable=C0111
        cont = {
            "indent": self._indent,
            "type": "Str",
            "str": node.s,
            "node": node,
            "value": node.s}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Name(self, node):  # pylint: disable=C0111
        cont = {
            "indent": self._indent,
            "type": "Name",
            "str": node.id,
            "node": node,
            "id": node.id,
            "ctx": node.ctx}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Module(self, node):  # pylint: disable=C0111
        cont = {
            "indent": self._indent,
            "type": "Module",
            "str": "",
            "body": node.body,
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_FunctionDef(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "FunctionDef", "str": node.name, "name": node.name, "body": node.body,
                "node": node, "returns": node.returns}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_List(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "List",
                "str": "", "elts": node.elts,
                "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_arguments(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "arguments", "str": "",
                "node": node, "args": node.args}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_arg(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "arg", "str": node.arg,
                "node": node,
                "arg": node.arg, "annotation": node.annotation}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Assign(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Assign", "str": "", "node": node,
                "targets": node.targets, "value": node.value}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Store(self, node):  # pylint: disable=C0111
        #cont = { "indent":self._indent, "type": "Store", "str": "" }
        # self.push(cont)
        cont = {}
        return self.generic_visit_args(node, cont)

    def visit_Call(self, node):  # pylint: disable=C0111
        if "attr" in node.func.__dict__:
            cont = {"indent": self._indent, "type": "Call", "str": node.func.attr,
                    "node": node, "func": node.func}
        else:
            cont = {"indent": self._indent, "type": "Call", "str": node.func.id,
                    "node": node, "func": node.func}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Attribute(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Attribute", "str": node.attr,
                "node": node, "value": node.value, "ctx": node.ctx, "attr": node.attr}
        self.push(cont)
        # last = len(self._rows)
        res = self.generic_visit_args(node, cont)

        if len(cont["children"]) > 0:
            fir = cont["children"][0]
            if fir["type"] == "Name":
                parent = fir["node"].id
                cont["str"] = "{0}.{1}".format(parent, cont["str"])
                cont["children"][0]["remove"] = True
        return res

    def visit_Load(self, node):  # pylint: disable=C0111
        cont = {}
        return self.generic_visit_args(node, cont)

    def visit_keyword(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "keyword", "str": "{0}".format(node.arg),
                "node": node, "arg": node.arg, "value": node.value}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_BinOp(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "BinOp",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Div(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Div",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Sub(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Sub",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_USub(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Sub",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Add(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Add",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Pow(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Pow",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Mult(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Mult",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_MatMult(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "MatMult",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Compare(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Compare",
                "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Gt(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Gt", "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Lt(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Lt", "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_UnaryOp(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent,
                "type": "UnaryOp", "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Num(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Num",
                "node": node, "str": "{0}".format(node.n),
                'n': node.n}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Return(self, node):  # pylint: disable=C0111
        cont = {"indent": self._indent, "type": "Return", "node": node, "str": "",
                'value': node.value}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_NameConstant(self, node):
        """
        A name.
        """
        if node.value is None:
            cont = {"indent": self._indent, "type": "Cst",
                    "node": node, "str": "None",
                    'n': None}
            self.push(cont)
            return self.generic_visit_args(node, cont)
        return self.visit_(node)  # pragma: no cover



"""
@file
@brief Inspired from skl2onnx, handles two backends.
"""
import numpy
from ...tools.asv_options_helper import get_opset_number_from_onnx
from .utils_backend_onnxruntime import _capture_output


from .tests_helper import (  # noqa
    binary_array_to_string,
    dump_data_and_model,
    dump_one_class_classification,
    dump_binary_classification,
    dump_multilabel_classification,
    dump_multiple_classification,
    dump_multiple_regression,
    dump_single_regression,
    convert_model,
    fit_classification_model,
    fit_classification_model_simple,
    fit_multilabel_classification_model,
    fit_regression_model)


def create_tensor(N, C, H=None, W=None):
    "Creates a tensor."
    if H is None and W is None:
        return numpy.random.rand(N, C).astype(numpy.float32, copy=False)  # pylint: disable=E1101
    elif H is not None and W is not None:
        return numpy.random.rand(N, C, H, W).astype(numpy.float32, copy=False)  # pylint: disable=E1101
    raise ValueError(  # pragma no cover
        'This function only produce 2-D or 4-D tensor.')


def _get_ir_version(opv):
    if opv >= 12:
        return 7
    if opv >= 11:  # pragma no cover
        return 6
    if opv >= 10:  # pragma no cover
        return 5
    if opv >= 9:  # pragma no cover
        return 4
    if opv >= 8:  # pragma no cover
        return 4
    return 3  # pragma no cover


TARGET_OPSET = get_opset_number_from_onnx()
TARGET_IR = _get_ir_version(TARGET_OPSET)



"""
@file
@brief Functions to validate converted models and runtime.
"""

from .validate import enumerate_validated_operator_opsets
from .validate_benchmark_replay import enumerate_benchmark_replay
from .validate_summary import summary_report
from .validate_helper import sklearn_operators



"""
@file
@brief Validates runtime for many :scikit-learn: operators.
The submodule relies on :epkg:`onnxconverter_common`,
:epkg:`sklearn-onnx`.
"""
import numpy
from sklearn.base import (
    ClusterMixin, BiclusterMixin, OutlierMixin,
    RegressorMixin, ClassifierMixin)
from sklearn.calibration import CalibratedClassifierCV
from sklearn.cross_decomposition import PLSSVD
from sklearn.datasets import load_iris
from sklearn.decomposition import LatentDirichletAllocation, NMF
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from sklearn.ensemble import (
    AdaBoostRegressor, GradientBoostingRegressor, AdaBoostClassifier,
    BaggingClassifier, VotingClassifier, GradientBoostingClassifier,
    RandomForestClassifier)
try:
    from sklearn.ensemble import StackingClassifier, StackingRegressor
except ImportError:  # pragma: no cover
    # new in 0.22
    StackingClassifier, StackingRegressor = None, None
from sklearn.feature_extraction import DictVectorizer, FeatureHasher
from sklearn.feature_extraction.text import (
    CountVectorizer, TfidfVectorizer, TfidfTransformer)
from sklearn.experimental import enable_hist_gradient_boosting  # pylint: disable=W0611
from sklearn.ensemble import (
    HistGradientBoostingRegressor,
    HistGradientBoostingClassifier)
from sklearn.feature_selection import (
    RFE, RFECV, GenericUnivariateSelect,
    SelectPercentile, SelectFwe, SelectKBest,
    SelectFdr, SelectFpr, SelectFromModel)
from sklearn.gaussian_process import GaussianProcessClassifier, GaussianProcessRegressor
from sklearn.isotonic import IsotonicRegression
from sklearn.linear_model import (
    ARDRegression, ElasticNetCV,
    LarsCV, LassoCV, LassoLarsCV, LassoLarsIC,
    SGDRegressor, OrthogonalMatchingPursuitCV,
    TheilSenRegressor, BayesianRidge, MultiTaskElasticNet,
    MultiTaskElasticNetCV, MultiTaskLassoCV, MultiTaskLasso,
    PassiveAggressiveClassifier, RidgeClassifier,
    RidgeClassifierCV, PassiveAggressiveRegressor,
    HuberRegressor, LogisticRegression, SGDClassifier,
    LogisticRegressionCV, Perceptron)
from sklearn.mixture._base import BaseMixture
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.multiclass import (
    OneVsRestClassifier, OneVsOneClassifier, OutputCodeClassifier)
from sklearn.multioutput import MultiOutputRegressor, MultiOutputClassifier
from sklearn.naive_bayes import BernoulliNB, GaussianNB, MultinomialNB, ComplementNB
from sklearn.neighbors import (
    NearestCentroid, RadiusNeighborsClassifier,
    NeighborhoodComponentsAnalysis)
from sklearn.preprocessing import (
    LabelBinarizer, LabelEncoder,
    OneHotEncoder, PowerTransformer)
from sklearn.semi_supervised import LabelPropagation, LabelSpreading
from sklearn.svm import LinearSVC, LinearSVR, NuSVR, SVR, SVC, NuSVC
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier, ExtraTreeClassifier
from sklearn.utils import shuffle
from skl2onnx.common.data_types import (
    FloatTensorType, DoubleTensorType, StringTensorType, DictionaryType)
from ._validate_problems_helper import (
    _noshapevar, _1d_problem, text_alpha_num)


def _modify_dimension(X, n_features, seed=19):
    """
    Modifies the number of features to increase
    or reduce the number of features.

    @param      X           features matrix
    @param      n_features  number of features
    @param      seed        random seed (to get the same
                            dataset at each call)
    @return                 new featurs matrix
    """
    if n_features is None or n_features == X.shape[1]:
        return X
    if n_features < X.shape[1]:
        return X[:, :n_features]
    rstate = numpy.random.RandomState(seed)  # pylint: disable=E1101
    res = numpy.empty((X.shape[0], n_features), dtype=X.dtype)
    res[:, :X.shape[1]] = X[:, :]
    div = max((n_features // X.shape[1]) + 1, 2)
    for i in range(X.shape[1], res.shape[1]):
        j = i % X.shape[1]
        col = X[:, j]
        if X.dtype in (numpy.float32, numpy.float64):
            sigma = numpy.var(col) ** 0.5
            rnd = rstate.randn(len(col)) * sigma / div
            col2 = col + rnd
            res[:, j] -= col2 / div
            res[:, i] = col2
        elif X.dtype in (numpy.int32, numpy.int64):
            perm = rstate.permutation(col)
            h = rstate.randint(0, div) % X.shape[0]
            col2 = col.copy()
            col2[h::div] = perm[h::div]  # pylint: disable=E1136
            res[:, i] = col2
            h = (h + 1) % X.shape[0]
            res[h, j] = perm[h]  # pylint: disable=E1136
        else:  # pragma: no cover
            raise NotImplementedError(  # pragma: no cover
                "Unable to add noise to a feature for this type {}".format(X.dtype))
    return res


###########
# datasets
###########


def _problem_for_predictor_binary_classification(
        dtype=numpy.float32, n_features=None, add_nan=False):
    """
    Returns *X, y, intial_types, method, node name, X runtime* for a
    binary classification problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    y[y == 2] = 1
    if add_nan:
        rows = numpy.random.randint(0, X.shape[0] - 1, X.shape[0] // 3)
        cols = numpy.random.randint(0, X.shape[1] - 1, X.shape[0] // 3)
        X[rows, cols] = numpy.nan
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, y, [('X', X[:1].astype(dtype))],
            'predict_proba', 1, X.astype(dtype))


def _problem_for_predictor_multi_classification(dtype=numpy.float32, n_features=None):
    """
    Returns *X, y, intial_types, method, node name, X runtime* for a
    m-cl classification problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, y, [('X', X[:1].astype(dtype))],
            'predict_proba', 1, X.astype(dtype))


def _problem_for_mixture(dtype=numpy.float32, n_features=None):
    """
    Returns *X, y, intial_types, method, node name, X runtime* for a
    m-cl classification problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, None, [('X', X[:1].astype(dtype))],
            'predict_proba', 1, X.astype(dtype))


def _problem_for_predictor_multi_classification_label(dtype=numpy.float32, n_features=None):
    """
    Returns *X, y, intial_types, method, node name, X runtime* for a
    m-cl classification problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    y2 = numpy.zeros((y.shape[0], 3), dtype=numpy.int64)
    for i, _ in enumerate(y):
        y2[i, _] = 1
    for i in range(0, y.shape[0], 5):
        y2[i, (y[i] + 1) % 3] = 1
    X = X.astype(dtype)
    y2 = y2.astype(numpy.int64)
    return (X, y2, [('X', X[:1].astype(dtype))],
            'predict_proba', 1, X.astype(dtype))


def _problem_for_predictor_regression(many_output=False, options=None,
                                      n_features=None, nbrows=None,
                                      dtype=numpy.float32, add_nan=False,
                                      **kwargs):
    """
    Returns *X, y, intial_types, method, name, X runtime* for a
    regression problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target + numpy.arange(len(data.target)) / 100
    meth = 'predict' if kwargs is None else ('predict', kwargs)
    itt = [('X', X[:1].astype(dtype))]
    if n_features is not None:
        X = X[:, :n_features]
        itt = [('X', X[:1].astype(dtype))]
    if nbrows is not None:
        X = X[:nbrows, :]
        y = y[:nbrows]
        itt = [('X', X[:1].astype(dtype))]
    if options is not None:
        itt = itt, options
    if add_nan:
        rows = numpy.random.randint(0, X.shape[0] - 1, X.shape[0] // 3)
        cols = numpy.random.randint(0, X.shape[1] - 1, X.shape[0] // 3)
        X[rows, cols] = numpy.nan
    X = X.astype(dtype)
    y = y.astype(dtype)
    return (X, y, itt,
            meth, 'all' if many_output else 0, X.astype(dtype))


def _problem_for_predictor_multi_regression(many_output=False, options=None,
                                            n_features=None, nbrows=None,
                                            dtype=numpy.float32, **kwargs):
    """
    Returns *X, y, intial_types, method, name, X runtime* for a
    mregression problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target.astype(float) + numpy.arange(len(data.target)) / 100
    meth = 'predict' if kwargs is None else ('predict', kwargs)
    itt = [('X', X[:1].astype(dtype))]
    if n_features is not None:
        X = X[:, :n_features]
        itt = [('X', X[:1].astype(dtype))]
    if nbrows is not None:
        X = X[:nbrows, :]
        y = y[:nbrows]
        itt = [('X', X[:1].astype(dtype))]
    if options is not None:
        itt = itt, options
    y2 = numpy.empty((y.shape[0], 2))
    y2[:, 0] = y
    y2[:, 1] = y + 0.5
    X = X.astype(dtype)
    y2 = y2.astype(dtype)
    return (X, y2, itt,
            meth, 'all' if many_output else 0, X.astype(dtype))


def _problem_for_numerical_transform(dtype=numpy.float32, n_features=None):
    """
    Returns *X, intial_types, method, name, X runtime* for a
    transformation problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    X = X.astype(dtype)
    return (X, None, [('X', X[:1].astype(dtype))],
            'transform', 0, X.astype(dtype=numpy.float32))


def _problem_for_numerical_transform_positive(dtype=numpy.float32, n_features=None):
    """
    Returns *X, intial_types, method, name, X runtime* for a
    transformation problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*data.data.shape) / 3
    X = numpy.abs(data.data + rnd)
    X = _modify_dimension(X, n_features)
    X = X.astype(dtype)
    return (X, None, [('X', X[:1].astype(dtype))],
            'transform', 0, X.astype(dtype=numpy.float32))


def _problem_for_numerical_trainable_transform(dtype=numpy.float32, n_features=None):
    """
    Returns *X, intial_types, method, name, X runtime* for a
    transformation problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target + numpy.arange(len(data.target)) / 100
    X = X.astype(dtype)
    y = y.astype(dtype)
    return (X, y, [('X', X[:1].astype(dtype))],
            'transform', 0, X.astype(dtype))


def _problem_for_numerical_trainable_transform_cl(dtype=numpy.float32, n_features=None):
    """
    Returns *X, intial_types, method, name, X runtime* for a
    transformation problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, y, [('X', X[:1].astype(dtype))],
            'transform', 0, X.astype(dtype))


def _problem_for_clustering(dtype=numpy.float32, n_features=None):
    """
    Returns *X, intial_types, method, name, X runtime* for a
    clustering problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    X = X.astype(dtype)
    return (X, None, [('X', X[:1].astype(dtype))],
            'predict', 0, X.astype(dtype))


def _problem_for_clustering_scores(dtype=numpy.float32, n_features=None):
    """
    Returns *X, intial_types, method, name, X runtime* for a
    clustering problem, the score part, not the cluster.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    X = X.astype(dtype)
    return (X, None, [('X', X[:1].astype(dtype))],
            'transform', 1, X.astype(dtype))


def _problem_for_outlier(dtype=numpy.float32, n_features=None):
    """
    Returns *X, intial_types, method, name, X runtime* for a
    transformation problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    X = X.astype(dtype)
    return (X, None, [('X', X[:1].astype(dtype))],
            'predict', 0, X.astype(dtype))


def _problem_for_numerical_scoring(dtype=numpy.float32, n_features=None):
    """
    Returns *X, y, intial_types, method, name, X runtime* for a
    scoring problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    y = data.target.astype(dtype) + numpy.arange(len(data.target)) / 100
    y /= numpy.max(y)
    X = X.astype(dtype)
    y = y.astype(dtype)
    return (X, y, [('X', X[:1].astype(dtype))],
            'score', 0, X.astype(dtype))


def _problem_for_clnoproba(dtype=numpy.float32, n_features=None):
    """
    Returns *X, y, intial_types, method, name, X runtime* for a
    scoring problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, y, [('X', X[:1].astype(dtype))],
            'predict', 0, X.astype(dtype))


def _problem_for_clnoproba_binary(dtype=numpy.float32, n_features=None, add_nan=False):
    """
    Returns *X, y, intial_types, method, name, X runtime* for a
    scoring problem. Binary classification.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    y[y == 2] = 1
    if add_nan:
        rows = numpy.random.randint(0, X.shape[0] - 1, X.shape[0] // 3)
        cols = numpy.random.randint(0, X.shape[1] - 1, X.shape[0] // 3)
        X[rows, cols] = numpy.nan
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, y, [('X', X[:1].astype(dtype))],
            'predict', 0, X.astype(dtype))


def _problem_for_cl_decision_function(dtype=numpy.float32, n_features=None):
    """
    Returns *X, y, intial_types, method, name, X runtime* for a
    scoring problem.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, y, [('X', X[:1].astype(dtype))],
            'decision_function', 1, X.astype(dtype))


def _problem_for_cl_decision_function_binary(dtype=numpy.float32, n_features=None):
    """
    Returns *X, y, intial_types, method, name, X runtime* for a
    scoring problem. Binary classification.
    It is based on Iris dataset.
    """
    data = load_iris()
    X = data.data
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*X.shape) / 3
    X += rnd
    X = _modify_dimension(X, n_features)
    y = data.target
    y[y == 2] = 1
    X = X.astype(dtype)
    y = y.astype(numpy.int64)
    return (X, y, [('X', X[:1].astype(dtype))],
            'decision_function', 1, X.astype(dtype))


def _problem_for_label_encoder(dtype=numpy.int64, n_features=None):
    """
    Returns a problem for the :epkg:`sklearn:preprocessing:LabelEncoder`.
    """
    data = load_iris()
    # X = data.data
    y = data.target.astype(dtype)
    itt = [('X', y[:1].astype(dtype))]
    y = y.astype(dtype)
    return (y, None, itt, 'transform', 0, y)


def _problem_for_dict_vectorizer(dtype=numpy.float32, n_features=None):
    """
    Returns a problem for the :epkg:`sklearn:feature_extraction:DictVectorizer`.
    """
    data = load_iris()
    # X = data.data
    y = data.target
    y2 = [{_: dtype(1000 + i)} for i, _ in enumerate(y)]
    y2[0][2] = -2
    cltype = FloatTensorType if dtype == numpy.float32 else DoubleTensorType
    itt = [("X", DictionaryType(StringTensorType([1]), cltype([1])))]
    y2 = numpy.array(y2)
    y = y.astype(numpy.int64)
    return (y2, y, itt, 'transform', 0, y2)


def _problem_for_tfidf_vectorizer(dtype=numpy.float32, n_features=None):
    """
    Returns a problem for the :epkg:`sklearn:feature_extraction:text:TfidfVectorizer`.
    """
    X = numpy.array([_[0] for _ in text_alpha_num])
    y = numpy.array([_[1] for _ in text_alpha_num], dtype=dtype)
    itt = [("X", StringTensorType([None]))]
    return (X, y, itt, 'transform', 0, X)


def _problem_for_tfidf_transformer(dtype=numpy.float32, n_features=None):
    """
    Returns a problem for the :epkg:`sklearn:feature_extraction:text:TfidfTransformer`.
    """
    X = numpy.array([_[0] for _ in text_alpha_num])
    y = numpy.array([_[1] for _ in text_alpha_num], dtype=dtype)
    X2 = CountVectorizer().fit_transform(X).astype(dtype)
    cltype = FloatTensorType if dtype == numpy.float32 else DoubleTensorType
    itt = [("X", cltype([None, X2.shape[1]]))]
    return (X2, y, itt, 'transform', 0, X2)


def _problem_for_feature_hasher(dtype=numpy.float32, n_features=None):
    """
    Returns a problem for the :epkg:`sklearn:feature_extraction:DictVectorizer`.
    """
    data = load_iris()
    # X = data.data
    y = data.target
    y2 = [{("cl%d" % _): dtype(1000 + i)} for i, _ in enumerate(y)]
    y2[0]["cl2"] = -2
    cltype = FloatTensorType if dtype == numpy.float32 else DoubleTensorType
    itt = [("X", DictionaryType(StringTensorType([1]), cltype([1])))]
    y2 = numpy.array(y2)
    return (y2, y, itt, 'transform', 0, y2)


def _problem_for_one_hot_encoder(dtype=numpy.float32, n_features=None):
    """
    Returns a problem for the :epkg:`sklearn:preprocessing:OneHotEncoder`.
    """
    data = load_iris()
    state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
    rnd = state.randn(*data.data.shape) / 3
    X = _modify_dimension(data.data + rnd, n_features)
    X = X.astype(numpy.int32).astype(dtype)
    y = data.target
    X, y = shuffle(X, y, random_state=1)
    itt = [('X', X[:1].astype(dtype))]
    return (X[:, :1], y, itt, 'transform', 0, X[:, :1].astype(dtype))


def find_suitable_problem(model):
    """
    Determines problems suitable for a given
    :epkg:`scikit-learn` operator. It may be

    * `b-cl`: binary classification
    * `m-cl`: m-cl classification
    * `m-label`: classification m-label
      (multiple labels possible at the same time)
    * `reg`: regression
    * `m-reg`: regression multi-output
    * `num-tr`: transform numerical features
    * `num-tr-pos`: transform numerical positive features
    * `scoring`: transform numerical features, target is usually needed
    * `outlier`: outlier prediction
    * `linearsvc`: classifier without *predict_proba*
    * `cluster`: similar to transform
    * `num+y-tr`: similar to transform with targets
    * `num+y-tr-cl`: similar to transform with classes
    * `num-tr-clu`: similar to cluster, but returns
        scores or distances instead of cluster
    * `key-col`: list of dictionaries
    * `text-col`: one column of text

    Suffix `nofit` indicates the predictions happens
    without the model being fitted. This is the case
    for :epkg:`sklearn:gaussian_process:GaussianProcessRegressor`.
    The suffix `-cov` indicates the method `predict` was called
    with parameter ``return_cov=True``, `-std` tells
    method `predict` was called with parameter ``return_std=True``.
    The suffix ``-NSV`` creates an input variable
    like the following ``[('X', FloatTensorType([None, None]))]``.
    That's a way to bypass :epkg:`onnxruntime` shape checking
    as one part of the graph is designed to handle any
    kind of dimensions but apparently, if the input shape is
    precise, every part of the graph has to be precise. The strings
    used variables which means it is at the same time precise
    and unprecise. Suffix ``'-64'`` means the model will
    do double computations. Suffix ``-nop`` means the classifier
    does not implement method *predict_proba*. Suffix ``-1d``
    means a one dimension problem (one feature). Suffix ``-dec``
    checks method `decision_function`.

    The following script gives the list of :epkg:`scikit-learn`
    models and the problem they can be fitted on.

    .. runpython::
        :showcode:
        :rst:

        from mlprodict.onnxrt.validate.validate import (
            sklearn_operators, find_suitable_problem)
        from pyquickhelper.pandashelper import df2rst
        from pandas import DataFrame
        res = sklearn_operators()
        rows = []
        for model in res[:20]:
            name = model['name']
            row = dict(name=name)
            try:
                prob = find_suitable_problem(model['cl'])
                if prob is None:
                    continue
                for p in prob:
                    row[p] = 'X'
            except RuntimeError:
                pass
            rows.append(row)
        df = DataFrame(rows).set_index('name')
        df = df.sort_index()
        print(df2rst(df, index=True))

    The list is truncated. The full list can be found at
    :ref:`l-model-problem-list`.
    """
    from ...onnx_conv.validate_scenarios import find_suitable_problem as ext_find_suitable_problem

    def _internal(model):  # pylint: disable=R0911

        # checks that this model is not overwritten by this module
        ext = ext_find_suitable_problem(model)
        if ext is not None:
            return ext

        # Exceptions
        if model in {GaussianProcessRegressor}:
            # m-reg causes MemoryError on some machine.
            return ['~b-reg-NF-64',  # '~m-reg-NF-64',
                    '~b-reg-NF-cov-64',  # '~m-reg-NF-cov-64',
                    '~b-reg-NF-std-64',  # '~m-reg-NF-std-64',
                    '~b-reg-NSV-64',  # '~m-reg-NSV-64',
                    '~b-reg-cov-64',  # '~m-reg-cov-64',
                    '~b-reg-std-NSV-64',  # '~m-reg-std-NSV-64',
                    'b-reg', '~b-reg-64',  # 'm-reg'
                    ]

        if model in {DictVectorizer}:
            return ['key-int-col']

        if model in {TfidfVectorizer, CountVectorizer}:
            return ['text-col']

        if model in {TfidfTransformer}:
            return ['bow']

        if model in {FeatureHasher}:
            return ['key-str-col']

        if model in {OneHotEncoder}:
            return ['one-hot']

        if model in {LabelBinarizer, LabelEncoder}:
            return ['int-col']

        if model in {NuSVC, SVC, SGDClassifier,
                     HistGradientBoostingClassifier}:
            return ['b-cl', 'm-cl', '~b-cl-64', '~b-cl-nan']

        if model in {GaussianProcessClassifier}:
            return ['b-cl', 'm-cl', '~b-cl-64']

        if model in {BaggingClassifier, BernoulliNB, CalibratedClassifierCV,
                     ComplementNB, GaussianNB,
                     GradientBoostingClassifier, LabelPropagation, LabelSpreading,
                     LinearDiscriminantAnalysis, LogisticRegressionCV,
                     MultinomialNB, QuadraticDiscriminantAnalysis,
                     RandomizedSearchCV}:
            return ['b-cl', 'm-cl']

        if model in {Perceptron}:
            return ['~b-cl-nop', '~m-cl-nop', '~b-cl-dec', '~m-cl-dec']

        if model in {AdaBoostRegressor}:
            return ['b-reg', '~b-reg-64']

        if model in {HistGradientBoostingRegressor}:
            return ['b-reg', '~b-reg-64', '~b-reg-nan', '~b-reg-nan-64']

        if model in {LinearSVC, NearestCentroid}:
            return ['~b-cl-nop', '~b-cl-nop-64']

        if model in {RFE, RFECV}:
            return ['num+y-tr']

        if model in {GridSearchCV}:
            return ['b-cl', 'm-cl',
                    'b-reg', 'm-reg',
                    '~b-reg-64', '~b-cl-64',
                    'cluster', 'outlier', '~m-label']

        if model in {VotingClassifier}:
            return ['b-cl', 'm-cl']

        if StackingClassifier is not None and model in {StackingClassifier}:
            return ['b-cl']

        if StackingRegressor is not None and model in {StackingRegressor}:
            return ['b-reg']

        # specific scenarios
        if model in {IsotonicRegression}:
            return ['~num+y-tr-1d', '~b-reg-1d']

        if model in {ARDRegression, BayesianRidge, ElasticNetCV,
                     GradientBoostingRegressor,
                     LarsCV, LassoCV, LassoLarsCV, LassoLarsIC,
                     LinearSVR, NuSVR, OrthogonalMatchingPursuitCV,
                     PassiveAggressiveRegressor, SGDRegressor,
                     TheilSenRegressor, HuberRegressor, SVR}:
            return ['b-reg', '~b-reg-64']

        if model in {MultiOutputClassifier}:
            return ['m-cl', '~m-label']

        if model in {MultiOutputRegressor, MultiTaskElasticNet,
                     MultiTaskElasticNetCV, MultiTaskLassoCV,
                     MultiTaskLasso}:
            return ['m-reg']

        if model in {OneVsOneClassifier, OutputCodeClassifier,
                     PassiveAggressiveClassifier, RadiusNeighborsClassifier}:
            return ['~b-cl-nop', '~m-cl-nop']

        if model in {RidgeClassifier, RidgeClassifierCV}:
            return ['~b-cl-nop', '~m-cl-nop', '~m-label']

        # trainable transform
        if model in {GenericUnivariateSelect,
                     NeighborhoodComponentsAnalysis,
                     PLSSVD, SelectKBest,
                     SelectPercentile, SelectFromModel}:
            return ["num+y-tr"]

        if model in {SelectFwe, SelectFdr, SelectFpr}:
            return ["num+y-tr-cl"]

        # no m-label
        if model in {AdaBoostClassifier}:
            return ['b-cl', '~b-cl-64', 'm-cl']

        if model in {LogisticRegression}:
            return ['b-cl', '~b-cl-64', 'm-cl', '~b-cl-dec', '~m-cl-dec']

        if model in {RandomForestClassifier}:
            return ['b-cl', '~b-cl-64', 'm-cl', '~m-label']

        if model in {DecisionTreeClassifier, ExtraTreeClassifier}:
            return ['b-cl', '~b-cl-64', 'm-cl', '~b-cl-f100', '~m-label']

        if model in {DecisionTreeRegressor}:
            return ['b-reg', 'm-reg', '~b-reg-64', '~m-reg-64', '~b-reg-f100']

        if model in {LatentDirichletAllocation, NMF, PowerTransformer}:
            return ['num-tr-pos']

        if hasattr(model, 'predict'):
            if "Classifier" in str(model):
                return ['b-cl', '~b-cl-64', 'm-cl', '~m-label']
            elif "Regressor" in str(model):
                return ['b-reg', 'm-reg', '~b-reg-64', '~m-reg-64']

        # Generic case.
        res = []
        if hasattr(model, 'transform'):
            if issubclass(model, (RegressorMixin, ClassifierMixin)):
                res.extend(['num+y-tr'])
            elif issubclass(model, (ClusterMixin, BiclusterMixin)):
                res.extend(['~num-tr-clu', '~num-tr-clu-64'])
            else:
                res.extend(['num-tr'])

        if hasattr(model, 'predict') and issubclass(model, (ClusterMixin, BiclusterMixin)):
            res.extend(['cluster', '~b-clu-64'])

        if issubclass(model, (OutlierMixin)):
            res.extend(['outlier'])

        if issubclass(model, ClassifierMixin):
            if model is OneVsRestClassifier:
                return ['m-cl', '~m-label']
            res.extend(['b-cl', '~b-cl-64', 'm-cl', '~m-label'])
        if issubclass(model, RegressorMixin):
            res.extend(['b-reg', 'm-reg', '~b-reg-64', '~m-reg-64'])
        if issubclass(model, BaseMixture):
            res.extend(['mix', '~mix-64'])

        if len(res) > 0:
            return res

        raise RuntimeError("Unable to find problem for model '{}' - {}."
                           "".format(model.__name__, model.__bases__))

    res = _internal(model)
    for r in res:
        if r not in _problems:
            raise ValueError(  # pragma: no cover
                "Unrecognized problem '{}' in\n{}".format(
                    r, "\n".join(sorted(_problems))))
    return res


_problems = {
    # standard
    "b-cl": _problem_for_predictor_binary_classification,
    "m-cl": _problem_for_predictor_multi_classification,
    "b-reg": _problem_for_predictor_regression,
    "m-reg": _problem_for_predictor_multi_regression,
    "num-tr": _problem_for_numerical_transform,
    "num-tr-pos": _problem_for_numerical_transform_positive,
    'outlier': _problem_for_outlier,
    'cluster': _problem_for_clustering,
    'num+y-tr': _problem_for_numerical_trainable_transform,
    'num+y-tr-cl': _problem_for_numerical_trainable_transform_cl,
    'mix': _problem_for_mixture,
    # others
    '~num-tr-clu': _problem_for_clustering_scores,
    "~m-label": _problem_for_predictor_multi_classification_label,
    "~scoring": _problem_for_numerical_scoring,
    '~b-cl-nop': _problem_for_clnoproba_binary,
    '~m-cl-nop': _problem_for_clnoproba,
    '~b-cl-dec': _problem_for_cl_decision_function_binary,
    '~m-cl-dec': _problem_for_cl_decision_function,
    # nan
    "~b-reg-nan": lambda n_features=None: _problem_for_predictor_regression(
        n_features=n_features, add_nan=True),
    "~b-reg-nan-64": lambda n_features=None: _problem_for_predictor_regression(
        dtype=numpy.float64, n_features=n_features, add_nan=True),
    "~b-cl-nan": lambda dtype=numpy.float32, n_features=None: _problem_for_predictor_binary_classification(
        dtype=dtype, n_features=n_features, add_nan=True),
    # 100 features
    "~b-reg-f100": lambda n_features=100: _problem_for_predictor_regression(
        n_features=n_features or 100),
    "~b-cl-f100": lambda n_features=100: _problem_for_predictor_binary_classification(
        n_features=n_features or 100),
    # 64
    "~b-cl-64": lambda n_features=None: _problem_for_predictor_binary_classification(
        dtype=numpy.float64, n_features=n_features),
    "~b-reg-64": lambda n_features=None: _problem_for_predictor_regression(
        dtype=numpy.float64, n_features=n_features),
    '~b-cl-nop-64': lambda n_features=None: _problem_for_clnoproba(
        dtype=numpy.float64, n_features=n_features),
    '~b-clu-64': lambda n_features=None: _problem_for_clustering(
        dtype=numpy.float64, n_features=n_features),
    '~b-cl-dec-64': lambda n_features=None: _problem_for_cl_decision_function_binary(
        dtype=numpy.float64, n_features=n_features),
    '~num-tr-clu-64': lambda n_features=None: _problem_for_clustering_scores(
        dtype=numpy.float64, n_features=n_features),
    "~m-reg-64": lambda n_features=None: _problem_for_predictor_multi_regression(
        dtype=numpy.float64, n_features=n_features),
    "~num-tr-64": lambda n_features=None: _problem_for_numerical_transform(
        dtype=numpy.float64, n_features=n_features),
    '~mix-64': lambda n_features=None: _problem_for_mixture(
        dtype=numpy.float64, n_features=n_features),
    #
    "~b-cl-NF": (lambda n_features=None: _problem_for_predictor_binary_classification(
        n_features=n_features) + (False, )),
    "~m-cl-NF": (lambda n_features=None: _problem_for_predictor_multi_classification(
        n_features=n_features) + (False, )),
    "~b-reg-NF": (lambda n_features=None: _problem_for_predictor_regression(
        n_features=n_features) + (False, )),
    "~m-reg-NF": (lambda n_features=None: _problem_for_predictor_multi_regression(
        n_features=n_features) + (False, )),
    #
    "~b-cl-NF-64": (lambda n_features=None: _problem_for_predictor_binary_classification(
        dtype=numpy.float64, n_features=n_features) + (False, )),
    "~m-cl-NF-64": (lambda n_features=None: _problem_for_predictor_multi_classification(
        dtype=numpy.float64, n_features=n_features) + (False, )),
    "~b-reg-NF-64": (lambda n_features=None: _problem_for_predictor_regression(
        dtype=numpy.float64, n_features=n_features) + (False, )),
    "~m-reg-NF-64": (lambda n_features=None: _problem_for_predictor_multi_regression(
        dtype=numpy.float64, n_features=n_features) + (False, )),
    # GaussianProcess
    "~b-reg-NF-cov-64": (lambda n_features=None: _problem_for_predictor_regression(
        True, options={GaussianProcessRegressor: {"return_cov": True}},
        return_cov=True, dtype=numpy.float64, n_features=n_features) + (False, )),
    "~m-reg-NF-cov-64": (lambda n_features=None: _problem_for_predictor_multi_regression(
        True, options={GaussianProcessRegressor: {"return_cov": True}},
        return_cov=True, dtype=numpy.float64, n_features=n_features) + (False, )),
    #
    "~b-reg-NF-std-64": (lambda n_features=None: _problem_for_predictor_regression(
        True, options={GaussianProcessRegressor: {"return_std": True}},
        return_std=True, dtype=numpy.float64, n_features=n_features) + (False, )),
    "~m-reg-NF-std-64": (lambda n_features=None: _problem_for_predictor_multi_regression(
        True, options={GaussianProcessRegressor: {"return_std": True}},
        return_std=True, dtype=numpy.float64, n_features=n_features) + (False, )),
    #
    "~b-reg-cov-64": (lambda n_features=None: _problem_for_predictor_regression(
        True, options={GaussianProcessRegressor: {"return_cov": True}},
        return_cov=True, dtype=numpy.float64, n_features=n_features)),
    "~m-reg-cov-64": (lambda n_features=None: _problem_for_predictor_multi_regression(
        True, options={GaussianProcessRegressor: {"return_cov": True}},
        return_cov=True, dtype=numpy.float64, n_features=n_features)),
    #
    "~reg-std-64": (lambda n_features=None: _problem_for_predictor_regression(
        True, options={GaussianProcessRegressor: {"return_std": True}},
        return_std=True, dtype=numpy.float64, n_features=n_features)),
    "~m-reg-std-64": (lambda n_features=None: _problem_for_predictor_multi_regression(
        True, options={GaussianProcessRegressor: {"return_std": True}},
        return_std=True, dtype=numpy.float64, n_features=n_features)),
    #
    '~b-reg-NSV-64': _noshapevar(lambda n_features=None: _problem_for_predictor_regression(
        dtype=numpy.float64, n_features=n_features)),
    '~m-reg-NSV-64': _noshapevar(lambda n_features=None: _problem_for_predictor_multi_regression(
        dtype=numpy.float64, n_features=n_features)),
    "~b-reg-std-NSV-64": (_noshapevar(lambda n_features=None: _problem_for_predictor_regression(
        True, options={GaussianProcessRegressor: {"return_std": True}},
        return_std=True, dtype=numpy.float64, n_features=n_features))),
    "~m-reg-std-NSV-64": (_noshapevar(lambda n_features=None: _problem_for_predictor_multi_regression(
        True, options={GaussianProcessRegressor: {"return_std": True}},
        return_std=True, dtype=numpy.float64, n_features=n_features))),
    # isotonic
    "~b-reg-1d": _1d_problem(_problem_for_predictor_regression),
    '~num+y-tr-1d': _1d_problem(_problem_for_numerical_trainable_transform),
    # text
    "key-int-col": _problem_for_dict_vectorizer,
    "key-str-col": _problem_for_feature_hasher,
    "int-col": _problem_for_label_encoder,
    "one-hot": _problem_for_one_hot_encoder,
    'text-col': _problem_for_tfidf_vectorizer,
    'bow': _problem_for_tfidf_transformer,
}



"""
@file
@brief Optimisation of :epkg:`ONNX` graphs.
"""
import copy
import hashlib
from onnx.helper import make_graph
from ._onnx_optimisation_common import (  # pylint: disable=E0611
    _rename_node_input,
    _rename_node_output,
    _apply_optimisation_on_graph,
    _apply_remove_node_fct_node
)


def _hash_obj_content(obj, max_size=1000):
    """
    Hash the content of an object.
    """
    m = hashlib.sha256()
    if hasattr(obj, 'op_type'):
        # An operator.
        m.update(obj.op_type.encode('ascii'))
        m.update(len(obj.output).to_bytes(8, byteorder='big'))
        for i in obj.input:
            m.update(i.encode('ascii'))
        if hasattr(obj, 'attribute'):
            for att in obj.attribute:
                m.update(att.name.encode('ascii'))
                m.update(_hash_obj_content(att))
    else:
        # An initializer.
        obj = copy.deepcopy(obj)
        obj.name = ""
        obj.doc_string = ""
        m.update(obj.SerializeToString())

    content = m.digest()
    if len(content) > max_size:
        content = content[:max_size]
    return content


def onnx_remove_node_redundant(onnx_model, recursive=True, debug_info=None,
                               max_hash_size=1000):
    """
    Removes redundant part of the graph. A redundant part is
    a set of nodes which takes the same inputs and produces
    the same outputs. It first starts by looking into duplicated
    initializers, then looks into nodes taking the same inputs
    and sharing the same type and parameters.

    @param      onnx_model      onnx model
    @param      recursive       looks into subgraphs
    @param      debug_info      debug information (private)
    @param      max_hash_size   limit the size of a hash used to detect
                                identical subgraphs
    @return                     new onnx _model
    """
    if debug_info is None:
        debug_info = [str(type(onnx_model)).split('.')[-1].strip("'>")]
    else:
        debug_info = debug_info + \
            [str(type(onnx_model)).split('.')[-1].strip("'>")]

    if hasattr(onnx_model, 'graph'):
        return _apply_optimisation_on_graph(
            onnx_remove_node_redundant, onnx_model,
            recursive=recursive, debug_info=debug_info,
            max_hash_size=max_hash_size)

    def _enumerate_rename_list_nodes_inputs(nodes, rename):
        for i, node in enumerate(nodes):
            if node is None:
                yield False, i, None
                continue
            if any(set(node.input) & set(rename)):
                yield True, i, _rename_node_input(node, rename)
                continue
            yield False, i, node

    graph = onnx_model

    # Detects duplicated initializers.
    hashes = {}
    names = []
    rename = {}
    for init in graph.initializer:
        hs = _hash_obj_content(init, max_size=max_hash_size)
        if hs in hashes:
            # Already seen.
            rename[init.name] = hashes[hs]  # pragma: no cover
        else:
            # New.
            hashes[hs] = init.name
            names.append(init.name)

    new_inits = [init for init in graph.initializer if init.name in set(names)]

    # Renames node inputs.
    new_nodes = []
    new_nodes = list(graph.node)
    new_nodes = list(
        _[2] for _ in _enumerate_rename_list_nodes_inputs(new_nodes, rename))

    # Detects duplicated operators.
    graph_outputs = set(o.name for o in graph.output)
    node_hashes = {}
    changed = 1
    replace = {}
    while changed > 0:
        changed = 0
        nnodes = len(new_nodes)
        for i in range(nnodes):
            if i in replace:
                # Already removed.
                continue
            node = new_nodes[i]
            hash = _hash_obj_content(node, max_size=max_hash_size)
            if hash in node_hashes:
                ni = node_hashes[hash]
                if ni == i:
                    continue
                replace[i] = ni
                changed += 1

                # Specifies what to rename.
                # One exception: the output is one of the graph output.
                rep = new_nodes[ni]
                for old, nn in zip(node.output, rep.output):
                    if old in graph_outputs:
                        rename[nn] = old
                        new_nodes[ni] = _rename_node_output(
                            new_nodes[ni], nn, old)
                    else:
                        rename[old] = nn

                # Renames inputs.
                new_new_nodes = []
                renew_index = set()
                for changed, ci, node in _enumerate_rename_list_nodes_inputs(new_nodes, rename):
                    if changed:
                        renew_index.add(ci)
                    new_new_nodes.append(node)
                new_nodes = new_new_nodes

                # Renews hashes.
                renew_hash = set(
                    k for k, v in node_hashes.items() if v in renew_index)
                for hs in renew_hash:
                    del node_hashes[hs]
                new_nodes[i] = None
            else:
                node_hashes[hash] = i

    if recursive:
        # Handles subgraphs.
        for i in range(len(new_nodes)):  # pylint: disable=C0200
            node = new_nodes[i]
            if node is None or not (node.attribute):  # pylint: disable=C0325
                continue
            new_nodes[i] = _apply_remove_node_fct_node(
                onnx_remove_node_redundant,
                node, recursive=True, debug_info=debug_info + [node.name])

    # Finally create the new graph.
    nodes = list(filter(lambda n: n is not None, new_nodes))
    graph = make_graph(nodes, onnx_model.name,
                       onnx_model.input, onnx_model.output,
                       new_inits)

    graph.value_info.extend(onnx_model.value_info)  # pylint: disable=E1101
    return graph



"""
@file
@brief Implements a class able to compute the predictions
from on an :epkg:`ONNX` model.
"""
from onnx import helper


def enumerate_model_node_outputs(model, add_node=False):
    """
    Enumerates all the nodes of a model.

    @param      model       :epkg:`ONNX` graph
    @param      add_node    if False, the function enumerates
                            all output names from every node, otherwise, it
                            enumerates tuple (output name, node)
    @return                 enumerator
    """
    if not hasattr(model, "graph"):
        raise TypeError(  # pragma: no cover
            "Parameter model is not an ONNX model but "
            "{}".format(type(model)))
    for node in model.graph.node:
        for out in node.output:
            yield (out, node) if add_node else out


def select_model_inputs_outputs(model, outputs=None, inputs=None):
    """
    Takes a model and changes its outputs.

    @param      model       :epkg:`ONNX` model
    @param      inputs      new inputs, same ones if None
    @param      outputs     new outputs, same ones if None
    @return                 modified model

    The function removes unneeded files.
    """
    if inputs is not None:
        raise NotImplementedError(  # pragma: no cover
            "Parameter inputs cannot be empty.")
    if outputs is None:
        raise RuntimeError(  # pragma: no cover
            "Parameter outputs cannot be None.")
    if not isinstance(outputs, list):
        outputs = [outputs]

    mark_var = {}
    for out in enumerate_model_node_outputs(model):
        mark_var[out] = 0
    for inp in model.graph.input:
        mark_var[inp.name] = 0
    for out in outputs:
        if out not in mark_var:
            raise ValueError(  # pragma: no cover
                "Output '{}' not found in model.".format(out))
        mark_var[out] = 1

    nodes = model.graph.node[::-1]
    mark_op = {}
    for node in nodes:
        mark_op[node.name] = 0

    # We mark all the nodes we need to keep.
    nb = 1
    while nb > 0:
        nb = 0
        for node in nodes:
            if mark_op[node.name] == 1:
                continue
            mod = False
            for out in node.output:
                if mark_var[out] == 1:
                    mark_op[node.name] = 1
                    mod = True
                    break
            if not mod:
                continue

            nb += 1
            for inp in node.input:
                if mark_var.get(inp, 0) == 1:
                    continue
                mark_var[inp] = 1
                nb += 1

    # All nodes verifies mark_op[node.name] == 1
    keep_nodes = [node for node in nodes if mark_op[node.name] == 1]

    var_out = []
    for out in outputs:
        value_info = helper.ValueInfoProto()
        value_info.name = out
        var_out.append(value_info)
    graph = helper.make_graph(keep_nodes, model.graph.name, model.graph.input,
                              var_out, model.graph.initializer)
    onnx_model = helper.make_model(graph)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:
        values = {p.key: p.value for p in model.metadata_props}
        helper.set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version

    if len(onnx_model.graph.input) != len(model.graph.input):  # pylint: disable=E1101
        raise RuntimeError(  # pragma: no cover
            "Input mismatch {} != {}".format(
                len(onnx_model.input), len(model.input)))  # pylint: disable=E1101
    return onnx_model



"""
@file
@brief Rewrites some of the converters implemented in
:epkg:`sklearn-onnx`.
"""
import numpy
from skl2onnx.operator_converters.decision_tree import (
    convert_sklearn_decision_tree_regressor,
    convert_sklearn_decision_tree_classifier)
from skl2onnx.operator_converters.gradient_boosting import (
    convert_sklearn_gradient_boosting_regressor,
    convert_sklearn_gradient_boosting_classifier)
from skl2onnx.operator_converters.random_forest import (
    convert_sklearn_random_forest_classifier,
    convert_sklearn_random_forest_regressor_converter)
from skl2onnx.common.data_types import guess_numpy_type


def _op_type_domain_regressor(dtype):
    """
    Defines *op_type* and *op_domain* based on `dtype`.
    """
    if dtype == numpy.float32:
        return 'TreeEnsembleRegressor', 'ai.onnx.ml', 1
    if dtype == numpy.float64:
        return 'TreeEnsembleRegressorDouble', 'mlprodict', 1
    raise RuntimeError(  # pragma: no cover
        "Unsupported dtype {}.".format(dtype))


def _op_type_domain_classifier(dtype):
    """
    Defines *op_type* and *op_domain* based on `dtype`.
    """
    if dtype == numpy.float32:
        return 'TreeEnsembleClassifier', 'ai.onnx.ml', 1
    if dtype == numpy.float64:
        return 'TreeEnsembleClassifierDouble', 'mlprodict', 1
    raise RuntimeError(  # pragma: no cover
        "Unsupported dtype {}.".format(dtype))


def new_convert_sklearn_decision_tree_classifier(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_classifier(dtype)
    convert_sklearn_decision_tree_classifier(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)


def new_convert_sklearn_decision_tree_regressor(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_regressor(dtype)
    convert_sklearn_decision_tree_regressor(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)


def new_convert_sklearn_gradient_boosting_classifier(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_classifier(dtype)
    convert_sklearn_gradient_boosting_classifier(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)


def new_convert_sklearn_gradient_boosting_regressor(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_regressor(dtype)
    convert_sklearn_gradient_boosting_regressor(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)


def new_convert_sklearn_random_forest_classifier(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_classifier(dtype)
    convert_sklearn_random_forest_classifier(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)


def new_convert_sklearn_random_forest_regressor(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_regressor(dtype)
    convert_sklearn_random_forest_regressor_converter(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)



"""
@file
@brief Actions definition.
"""
from .api_extension import AutoAction
from .gactions import MLAction, MLActionFunction


class MLModel(MLActionFunction):
    """
    Base class for every machine learned model
    """

    def __init__(self, input, output_names=None, name=None):
        """
        @param  name             a name which identifies the action
        @param  input           an action which produces the output result
        @param  output_names    names for the outputs
        """
        MLActionFunction.__init__(self, input, name=name)
        self.output_names = output_names

    @property
    def InputNames(self):
        """
        Returns the input names
        """
        vars = self.enumerate_variables()
        res = list(sorted(set(v.name_var for v in vars)))
        if len(res) == 0:
            raise ValueError(  # pragma: no cover
                "At least one variable must be defined.")
        return res

    @property
    def OutputNames(self):
        """
        Returns the output names
        """
        return self.output_names

    @AutoAction.cache
    def _export_json(self, hook=None, result_name=None):
        js = MLAction._export_json(self, hook=hook)
        js.update({"input_names": self.InputNames,
                   "output_names": self.OutputNames})
        return js

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            result_name = "pred"
        return MLActionFunction._export_c(self, hook=hook, result_name=result_name)



"""
@file
@brief Useful plots.
"""
import matplotlib.pyplot as plt
from ..onnxrt import OnnxInference


def plot_onnx(onx, ax=None, dpi=300, temp_dot=None, temp_img=None):
    """
    Plots an ONNX graph into graph.

    :param onx: ONNX object, @see cl OnnxInference
    :param ax: existing axes
    :param dpi: resolution
    :param temp_dot: temporary file,
        if None, a file is created and removed
    :param temp_img: temporary image,
        if None, a file is created and removed
    :return: axes
    """
    # delayed import
    from pyquickhelper.helpgen.graphviz_helper import plot_graphviz

    if ax is None:
        ax = plt.gca()
    if not isinstance(onx, OnnxInference):
        oinf = OnnxInference(onx, skip_run=True)
    else:
        oinf = onx
    dot = oinf.to_dot()
    plot_graphviz(dot, dpi=dpi, ax=ax, temp_dot=temp_dot, temp_img=temp_img)
    return ax



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject


class QuantizeLinear(OpRun):

    atts = {'axis': 1}
    python_inputs = ['*inputs']

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=QuantizeLinear.atts,
                       **options)

    def _run(self, *args):  # pylint: disable=W0221
        if len(args[1].shape) > 1:
            raise RuntimeError(  # pragma: no cover
                "Input 2 must be a vector or a number.")
        if len(args[1].shape) > 0:
            new_shape = [1 for s in args[0].shape]
            new_shape[self.axis] = len(args[1])
            x = args[0] / args[1].reshape(new_shape)
        else:
            x = args[0] / args[1]
        if len(args) > 2:
            dtype = args[2].dtype
            if len(args[1].shape) > 0:
                x += args[2].reshape(new_shape)
            else:
                x += args[2]
            numpy.around(x, 1, out=x)
            if dtype == numpy.uint8:
                numpy.clip(x, 0, 255, out=x)
            elif dtype == numpy.int8:
                numpy.clip(x, -128, 127, out=x)
            else:
                raise RuntimeError(  # pragma no cover
                    "Unexpected dtype for input 2 {}.".format(dtype))
            return (x.astype(dtype), )

        dtype = numpy.uint8
        numpy.around(x, 1, out=x)
        numpy.clip(x, 0, 255, out=x)
        return (x.astype(dtype), )

    def _infer_shapes(self, *args):  # pylint: disable=W0221
        if len(args) > 2:
            dtype = args[2].dtype
        else:
            dtype = numpy.uint8
        return (ShapeObject(args[0].shape, dtype=dtype), )



"""
@file
@brief Action definition.
"""
import numpy
from .api_extension import AutoAction
from .gtypes import (
    MLType, MLNumTypeFloat32, MLNumTypeFloat64,
    MLTensor, MLNumTypeInt32, MLNumTypeInt64, MLNumTypeBool)


class MLAction(AutoAction):
    """
    Base class for every action.
    """

    def __init__(self, inputs, output, name, children=None):
        """
        @param      inputs      type of inputs
        @param      output      output type
        @param      name        a name which identifies the action
        @param      children    actions used to compute this one
        """
        if not isinstance(inputs, list):
            raise TypeError(
                'inputs must be a list of MLType.')  # pragma: no cover
        for t in inputs:
            if not isinstance(t, MLType):
                raise TypeError(  # pragma: no cover
                    "Every input must be a MLType not '{0}'.".format(type(t)))
        if not isinstance(output, MLType):
            raise TypeError('output must be of MLType.')  # pragma: no cover
        self.inputs = inputs
        self.output = output
        self.name = name
        self.children = children if children is not None else []
        for child in self.children:
            if not isinstance(child, MLAction):  # pragma: no cover
                raise TypeError("All children must be of type MLAction")

    def execute(self, **kwargs):
        """
        Computes the action. Returns the output.
        """
        # It must be overwritten.
        self.children_results_ = [child.execute(
            **kwargs) for child in self.children]
        for v, tv in zip(self.children_results_, self.inputs):
            tv.validate(v)

    @property
    def ChildrenResults(self):
        """
        Return the last execution results.
        """
        return self.children_results_

    def enumerate_variables(self):
        """
        Enumerates all variables.
        """
        for child in self.children:
            for var in child.enumerate_variables():
                yield var

    def graph_execution(self):
        """
        Returns a formated string which retruns the outputs.
        """
        rows = []
        rows.append("-- BEGIN {0}  {3}                                             id={1} output={2}".format(
            self.name, id(self), self.output._cache, getattr(self, "comment", "")))
        for i, ch in enumerate(self.children):
            gr = ch.graph_execution()
            temp = ["    " + li for li in gr.split("\n")]
            temp[0] = "  {0}-".format(i) + temp[0][4:]
            rows.extend(temp)
        rows.append(
            "-- END {0} -- output={1}".format(self.name, self.output._cache))
        return "\n".join(rows)

    @AutoAction.cache
    def _export_json(self, hook=None, result_name=None):
        val = {"output": self.output._export_json()}
        if self.children:
            val["action"] = dict(name=self.name,
                                 variants=[c._export_json(hook=hook) for c in self.children])
        else:
            val["action"] = dict(name=self.name)
        if self.inputs:
            val["input"] = [i._export_json(hook=hook)
                            for i in self.inputs]
        return val

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError(
                "result_name must not be None")  # pragma: no cover
        rows = []
        rows.append("// {0}-{1} - children".format(id(self), self.name))
        names = []
        if self.children:
            for i, c in enumerate(self.children):
                rname = "{0}{1}{2}".format(
                    result_name, getattr(self, "cname", ""), i)
                dc = c._export_c(hook=hook, result_name=rname)
                if not dc['cache']:
                    rows.append(dc['code'])
                names.append(dc['result_name'])
        rows.append("// {0}-{1} - itself".format(id(self), self.name))
        res = "\n".join(rows)
        return {'code': res, 'result_name': result_name, 'child_names': names}


class MLActionCst(MLAction):
    """
    Constant
    """

    def __init__(self, cst, inout_type=None, comment=None):
        """
        @param  cst             constant
        @param  inout_type      type
        @param  comment         comment
        """
        if inout_type is None:
            inout_type = MLActionCst.guess_type(cst)
        MLAction.__init__(self, [], inout_type, "cst")
        inout_type.validate(cst)
        self.cst = cst
        self.comment = comment

    @staticmethod
    def guess_type(value):
        """
        Guesses a type given a value.
        """
        if isinstance(value, numpy.float32):
            return MLNumTypeFloat32()
        if isinstance(value, numpy.float64):
            return MLNumTypeFloat64()
        if isinstance(value, (int, numpy.int32)):
            return MLNumTypeInt32()
        if isinstance(value, (int, numpy.int64)):
            return MLNumTypeInt64()
        if isinstance(value, numpy.ndarray):
            a = numpy.zeros(1, value.dtype)
            t = MLActionCst.guess_type(a[0])
            return MLTensor(t, value.shape)
        raise NotImplementedError(  # pragma: no cover
            "Not implemented for type '{0}'".format(type(value)))

    def execute(self, **kwargs):
        MLAction.execute(self, **kwargs)
        return self.output.validate(self.cst)

    def graph_execution(self):
        if self.comment:
            return "cst: {0} = {1}".format(self.comment, self.cst)
        return "cst: {0}".format(self.cst)

    @AutoAction.cache
    def _export_json(self, hook=None, result_name=None):
        res = {"name": "cst",
               "value": self.output._format_value_json(self.cst, hook=hook)}
        if hasattr(self, "comment"):
            res["comment"] = self.comment
        return res

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError("result_name cannot be None.")  # pragma: no cover
        dc = self.output._export_c(hook='declare', result_name=result_name)
        res = "{0} = {1};".format(
            dc['code'], self.output._format_value_c(self.cst))
        if self.comment:
            res += " // {0}".format(self.comment)
        return {'code': res, 'result_name': result_name}


class MLActionVar(MLActionCst):
    """
    Variable. The constant is only needed to guess the
    variable type.
    """

    def __init__(self, value, name, inout_type=None):
        """
        @param  value       value
        @param  name        variable name
        @param  inout_type  type
        """
        MLActionCst.__init__(self, value, inout_type)
        self.name = "var"
        self.name_var = name

    def execute(self, **kwargs):
        MLAction.execute(self, **kwargs)
        if self.name_var not in kwargs:
            raise KeyError(  # pragma: no cover
                "Unable to find variable name '{0}'".format(self.name_var))
        return self.output.validate(kwargs[self.name_var])

    def enumerate_variables(self):
        """
        Enumerates itself.
        """
        yield self

    def graph_execution(self):
        return "var: {0} = {1} ({2})".format(self.name_var, self.name, self.output._cache)

    @AutoAction.cache
    def _export_json(self, hook=None, result_name=None):
        return {"name": "var", "value": self.name_var}

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError(  # pragma: no cover
                "result_name must not be None")
        dc = self.output._export_c(hook='typeref', result_name=result_name)
        res = "{0} = {1};".format(dc['code'], self.name_var)
        return {'code': res, 'result_name': result_name}


class MLActionFunctionCall(MLAction):
    """
    Any function call.
    """

    def __init__(self, name, output, *acts):
        """
        @param  name        function name
        @param  output      type
        @param  *acts       list of arguments
        """
        for act in acts:
            if not isinstance(act, MLAction):
                raise TypeError(  # pragma: no cover
                    "All element of acts must be MLAction not '{0}'.".format(type(act)))
        MLAction.__init__(self, [act.output for act in acts],
                          output, name, children=acts)
        self.cname = 'c'

    def _optional_parameters(self):
        """
        Returns additional parameters to add the function call.
        """
        return None

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError(
                "result_name must not be None")  # pragma: no cover
        dcf = MLAction._export_c(self, hook=hook, result_name=result_name)
        rows = [dcf['code']]
        fcall = ", ".join(dcf['child_names'])
        add = self._optional_parameters()  # pylint: disable=E1128
        if add is not None:
            fcall = ", ".join([fcall, add])
        dc = self.output._export_c(hook='declare', result_name=result_name)
        rows.append(dc['code'] + ";")
        ep = self.output._byref_c()
        type_list = "_".join(c.output.CTypeSingle for c in self.children)
        rows.append("{0}_{4}({3}{1}, {2});".format(
            self.name, result_name, fcall, ep, type_list))
        rows.append("// {0}-{1} - done".format(id(self), self.name))
        # Addition printf to debug the C++ code.
        # rows.append('printf("C++ {1} %f\\n", {0});'.format(result_name, self.name))
        res = {'code': "\n".join(rows), 'result_name': dcf['result_name']}
        return res


class MLActionBinary(MLAction):
    """
    Any binary operation.
    """

    def __init__(self, act1, act2, name):
        """
        @param  act1        first element
        @param  act2        second element
        @param  name        operator name
        """
        if not isinstance(act1, MLAction):
            raise TypeError("act1 must be MLAction.")  # pragma: no cover
        if not isinstance(act2, MLAction):
            raise TypeError("act2 must be MLAction.")  # pragma: no cover
        MLAction.__init__(self, [act1.output, act2.output], act2.output, name,
                          children=[act1, act2])

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError(
                "result_name must not be None")  # pragma: no cover
        dc = MLAction._export_c(self, hook=hook, result_name=result_name)
        rows = [dc['code']]
        dc2 = self.output._export_c(hook='type')
        op = "{2} {0} = {0}0 {1} {0}1;".format(
            result_name, self.name, dc2['code'])
        rows.append(op)
        rows.append("// {0}-{1} - done".format(id(self), self.name))
        return {'code': "\n".join(rows), 'result_name': result_name}


class MLActionUnary(MLAction):
    """
    Any binary operation.
    """

    def __init__(self, act1, name):
        """
        @param  act1        element
        @param  name        operator name
        """
        if not isinstance(act1, MLAction):
            raise TypeError("act1 must be MLAction.")  # pragma: no cover
        MLAction.__init__(self, [act1.output], act1.output, name,
                          children=[act1])

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError(  # pragma: no cover
                "result_name must not be None")
        dc = MLAction._export_c(self, hook=hook, result_name=result_name)
        rows = [dc['code']]
        op = "auto {0} = {1} {0}0;".format(result_name, self.name)
        rows.append(op)
        rows.append("// {0}-{1} - done".format(id(self), self.name))
        return {'code': "\n".join(rows), 'result_name': result_name}


class MLActionConcat(MLActionFunctionCall):
    """
    Concatenate number of arrays into an array.
    """

    def __init__(self, act1, act2):
        """
        @param  act1        first element
        @param  act2        second element
        """
        if not isinstance(act1, MLAction):
            raise TypeError("act1 must be MLAction.")  # pragma: no cover
        if not isinstance(act2, MLAction):
            raise TypeError("act2 must be MLAction.")  # pragma: no cover
        n1 = (1 if isinstance(act1.output, (MLNumTypeFloat32, MLNumTypeFloat64))
              else act1.output.dim[0])
        n2 = (1 if isinstance(act2.output, (MLNumTypeFloat32, MLNumTypeFloat64))
              else act2.output.dim[0])
        MLActionFunctionCall.__init__(self, "concat", MLTensor(
            act1.output.__class__(), (n1 + n2,)), act1, act2)

    def execute(self, **kwargs):
        """
        Concatenation
        """
        MLActionFunctionCall.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(numpy.array(res))


class MLActionCast(MLActionUnary):
    """
    Cast into another type.
    """

    def __init__(self, act1, new_type):
        """
        @param  act1        element
        @param  new_type    new type
        """
        MLActionUnary.__init__(self, act1, "cast")
        self.output = new_type

    def execute(self, **kwargs):
        MLActionUnary.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(self.output.cast(res[0]))

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        raise NotImplementedError(  # pragma: no cover
            "Not enough information to do it here.")


class MLActionIfElse(MLAction):
    """
    Addition
    """

    def __init__(self, cond, act1, act2, check_type=True, comment=None):
        """
        @param  cond        condition
        @param  act1        first action
        @param  ect2        second action
        @param  check_type  check ype
        @param  comment     comment
        """
        if not isinstance(act1, MLAction):
            raise TypeError("act1 must be MLAction.")  # pragma: no cover
        if not isinstance(act2, MLAction):
            raise TypeError("act2 must be MLAction.")  # pragma: no cover
        if not isinstance(cond, MLAction):
            raise TypeError("cond must be MLAction.")  # pragma: no cover
        if not isinstance(cond.output, MLNumTypeBool):
            raise TypeError(  # pragma: no cover
                "No boolean condition {0}".format(type(cond.output)))
        if check_type and type(act1.output) != type(act2.output):
            raise TypeError("Not the same input type {0} != {1}".format(  # pragma: no cover
                type(act1.output), type(act2.output)))
        MLAction.__init__(self, [cond.output, act1.output, act2.output], act2.output, "if",
                          children=[cond, act1, act2])
        self.comment = comment

    def execute(self, **kwargs):
        self.children_results_ = [
            self.children[0].execute(**kwargs), None, None]
        self.inputs[0].validate(self.children_results_[0])
        if self.children_results_[0]:
            self.children_results_[1] = self.children[1].execute(**kwargs)
            self.inputs[1].validate(self.children_results_[1])
            res = self.children_results_[1]
        else:
            self.children_results_[2] = self.children[2].execute(**kwargs)
            self.inputs[2].validate(self.children_results_[2])
            res = self.children_results_[2]
        return self.output.validate(res)

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError(
                "result_name must not be None")  # pragma: no cover
        dc = MLAction._export_c(self, hook=hook, result_name=result_name)
        rows = [dc['code']]
        dc2 = self.output._export_c(hook='type')
        op = "{1} {0} = {0}0 ? {0}1 : {0}2;".format(result_name, dc2['code'])
        rows.append(op)
        rows.append("// {0}-{1} - done".format(id(self), self.name))
        return {'code': "\n".join(rows), 'result_name': result_name}


class MLActionReturn(MLAction):
    """
    Returns a results.
    """

    def __init__(self, act):
        """
        @param  act     action to return
        """
        MLAction.__init__(self, [act.output],
                          act.output, "return", children=[act])

    def execute(self, **kwargs):
        MLAction.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(res[0])

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if len(self.children) != 1:
            raise ValueError(
                "Only one result can be returned.")  # pragma: no cover
        if result_name is None:
            raise ValueError(
                "result_name must not be None")  # pragma: no cover
        dc = self.children[0]._export_c(hook=hook, result_name=result_name)
        if not dc['cache']:
            code = dc['code']
        else:
            code = ''

        add = self.output._copy_c(
            result_name, result_name[:-1], hook="typeref")
        code += "\n" + add
        return {'code': code, 'result_name': result_name}


class MLActionFunction(MLActionUnary):
    """
    A function.
    """

    def __init__(self, act, name):
        """
        @param  act     action
        @param  name    name
        """
        if not isinstance(act, MLActionReturn):
            raise NotImplementedError(  # pragma: no cover
                "Last result must be MLActionReturn.")
        MLActionUnary.__init__(self, act, name)

    def execute(self, **kwargs):
        MLActionUnary.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(res[0])

    @AutoAction.cache
    def _export_c(self, hook=None, result_name=None):
        if result_name is None:
            raise ValueError(
                "result_name must not be None")  # pragma: no cover
        if len(self.children) != 1:
            raise ValueError(
                "The function must return one result.")  # pragma: no cover
        if result_name[-1] == '0':
            raise ValueError(  # pragma: no cover
                "result_name '{0}' cannot end with 0.".format(result_name))

        vars = {v.name: v for v in self.enumerate_variables()}
        vars = [_[1] for _ in list(sorted(vars.items()))]
        parameters = ", ".join("{0} {1}".format(
            v.output._export_c(hook='type')['code'], v.name_var) for v in vars)
        typename = self.children[0].output._export_c(
            hook='typeref', result_name=result_name)['code']
        signature = "int {1} ({0}, {2})".format(
            typename, self.name, parameters)
        dc = MLAction._export_c(self, hook=hook, result_name=result_name)
        code = dc['code']
        rows = [signature, "{"]
        rows.extend("    " + line for line in code.split("\n"))
        rows.extend(
            ['    return 0;', "    // {0}-{1} - done".format(id(self), self.name), '}'])
        return {'code': "\n".join(rows), 'result_name': result_name}



"""
@file
@brief Shape object.
"""
import numpy


class BaseDimensionShape:
    """
    Base class to @see cl DimensionObject,
    @see cl ShapeOperator, @see cl ShapeObject.
    """

    def to_string(self, use_x=True):
        """
        Converts the object into a string.
        """
        raise NotImplementedError()

    def evaluate(self, **kwargs):
        """
        Evaluates the object, reduces the expression
        to a number or a string.
        """
        raise NotImplementedError()  # pragma: no cover


class ShapeOperator(BaseDimensionShape):
    """
    Base class for all shapes operator.
    """

    def __init__(self, name, fct, fct_string, *args):
        """
        @param      name        display name of the operator
        @param      fct         function doing the operator
                                if argument are numeric
        @param      fct_string  function represented as a string
        @param      args        argument of the operator
        """
        self._name = name
        self._fct = fct
        self._fct_string = fct_string
        self._args = args
        for a in self._args:
            if not isinstance(a, DimensionObject):
                raise TypeError(
                    "All arguments must be of type DimensionObject not '{}'."
                    "".format(type(a)))

    def __repr__(self):
        """
        usual
        """
        return "{0}('{1}', {2}, '{2}', {3})".format(
            self.__class__.__name__, self._name,
            self._fct_string, self._args)

    def to_string(self, use_x=True):
        """
        Displays as a string.

        @return     a string
        """
        raise NotImplementedError(  # pragma: no cover
            "Operator '{}' does not implement 'to_string': {}.".format(
                self.__class__.__name__, repr(self)))

    def evaluate(self, **kwargs):
        """
        Evalutes the operator.

        @param  kwargs      value for the variables.
        @return             string or integer
        """
        args = []
        has_string = False
        for a in self._args:
            a = DimensionObject._same_(a)
            v = a.evaluate(**kwargs)
            if isinstance(v, str):
                has_string = True
            args.append(v)
        if has_string:
            res = self._evaluate_string_(args, **kwargs)
        else:
            try:
                res = self._fct(*args)
            except TypeError as e:
                raise RuntimeError(
                    "Unable to evaluate operator {} due to {}".format(repr(self), e)) from e
        return res

    def _evaluate_string_(self, args, **kwargs):
        """
        Evalutes the operator assuming some of them are still strings.

        @param  args        arguments extracted by method *evaluate*
        @param  kwargs      value for the variables.
        @return             string or integer
        """
        raise NotImplementedError(
            "This function must be overwritten.")  # pragma: no cover


class ShapeBinaryOperator(ShapeOperator):
    """
    Base class for shape binary operator.
    """

    def __init__(self, name, fct, fct_string, x, y):
        """
        @param      name        display name of the operator
        @param      fct         function doing the operator
                                if argument are numeric
        @param      fct_string  function represented as a string
        @param      x           first argument
        @param      y           second argument
        """
        ShapeOperator.__init__(self, name, fct, fct_string, x, y)
        if isinstance(x, tuple):
            raise TypeError('x cannot be a tuple')  # pragma: no cover
        if isinstance(y, tuple):
            raise TypeError('y cannot be a tuple')  # pragma: no cover

    def _to_string1(self, x, y):
        return DimensionObject(self._fct(x._dim, y._dim)).to_string()

    def _to_string2(self, x, y):
        return DimensionObject("{}{}{}".format(x._dim, self._name, y._dim)).to_string()

    def _to_string2b(self, x, y):
        return DimensionObject("({}){}({})".format(x._dim, self._name, y._dim)).to_string()

    def _to_string3(self, x):
        return DimensionObject("{}{}x".format(x._dim, self._name)).to_string()

    def to_string(self, use_x=True):
        """
        Applies binary operator to a dimension.

        @param      use_x   use `'x'` if dimension is unknown
        @return             a string
        """
        x, y = self._args  # pylint: disable=W0632
        if isinstance(x._dim, int):
            if isinstance(y, DimensionObject):
                if isinstance(y._dim, int):
                    return self._to_string1(x, y)
                if isinstance(y._dim, str):
                    return self._to_string2(x, y)
                if y._dim is None:
                    if use_x:
                        return self._to_string3(x)
                    return DimensionObject("{}{}DimensionObject()".format(
                        x._dim, self._name)).to_string()
                raise TypeError(  # pragma: no cover
                    "Unable to handle type '{}'.".format(type(y._dim)))
            raise TypeError(  # pragma: no cover
                "Unable to handle type '{}'.".format(type(y)))
        elif isinstance(x._dim, str):
            if isinstance(y._dim, int):
                return self._to_string2(x, y)
            if isinstance(y._dim, str):
                return self._to_string2b(x, y)
            raise TypeError(  # pragma: no cover
                "Unable to handle type '{}'.".format(type(y._dim)))
        raise TypeError(  # pragma: no cover
            "Unable to handle type '{}'.".format(type(x._dim)))

    def _evaluate_string_(self, args, **kwargs):
        """
        Evalutes the operator assuming some of them are still strings.

        @param  args        arguments extracted by method *evaluate*
        @param  kwargs      value for the variables.
        @return             string or integer
        """
        return self._name.join(map(lambda s: '({})'.format(s), args))


class ShapeBinaryFctOperator(ShapeBinaryOperator):
    """
    Base class for shape binary operator defined by a function.
    """

    def _to_string2(self, x, y):
        return DimensionObject("{}({},{})".format(self._name, x._dim, y._dim)).to_string()

    def _to_string2b(self, x, y):
        return DimensionObject("{}({},{})".format(self._name, x._dim, y._dim)).to_string()

    def _to_string3(self, x):
        return DimensionObject("{}({},x)".format(self._name, x._dim)).to_string()

    def _evaluate_string_(self, args, **kwargs):
        """
        Evalutes the operator assuming some of them are still strings.

        @param  args        arguments extracted by method *evaluate*
        @param  kwargs      value for the variables.
        @return             string or integer
        """
        return "{}({})".format(self._name, ",".join(map(str, args)))


class ShapeOperatorAdd(ShapeBinaryOperator):
    """
    Shape addition.
    """

    def __init__(self, x, y):
        ShapeBinaryOperator.__init__(
            self, '+', lambda a, b: a + b, 'lambda a, b: a + b', x, y)

    def __repr__(self):
        """
        Displays a string.

        @return             a string
        """
        return "{0}({1}, {2})".format(
            self.__class__.__name__, repr(self._args[0]), repr(self._args[1]))


class ShapeOperatorMul(ShapeBinaryOperator):
    """
    Shape multiplication.
    """

    def __init__(self, x, y):
        ShapeBinaryOperator.__init__(
            self, '*', lambda a, b: a * b, 'lambda a, b: a * b', x, y)

    def __repr__(self):
        """
        Displays a string.

        @return             a string
        """
        return "{0}({1}, {2})".format(
            self.__class__.__name__, repr(self._args[0]), repr(self._args[1]))


class ShapeOperatorGreater(ShapeBinaryOperator):
    """
    Shape comparison.
    """

    def __init__(self, x, y):
        ShapeBinaryOperator.__init__(
            self, '>', lambda a, b: a > b, 'lambda a, b: a > b', x, y)

    def __repr__(self):
        """
        Displays a string.

        @return             a string
        """
        return "{0}({1}, {2})".format(
            self.__class__.__name__, repr(self._args[0]), repr(self._args[1]))


class ShapeOperatorMax(ShapeBinaryFctOperator):
    """
    Best on each dimension.
    """

    def __init__(self, x, y):
        ShapeBinaryFctOperator.__init__(
            self, 'max', lambda a, b: max(a, b), 'max(a, b)', x, y)

    def __repr__(self):
        """
        Displays a string.

        @return             a string
        """
        return "{0}({1}, {2})".format(
            self.__class__.__name__, repr(self._args[0]), repr(self._args[1]))


class DimensionObject(BaseDimensionShape):
    """
    One dimension of a shape.
    """

    def __init__(self, obj):
        """
        @param  obj     int or @see cl DimensionObject or None to
                        specify something unknown
        """
        if obj is None or obj == 0 or obj == '?':
            self._dim = None
        elif isinstance(obj, (int, str, ShapeOperator, DimensionObject,
                              numpy.int32, numpy.int64)):
            self._dim = obj
        else:
            raise TypeError("Unexpected type for obj: {}".format(type(obj)))

    @property
    def dim(self):
        """
        Returns the dimension.
        """
        return self._dim

    def __repr__(self):
        """
        usual
        """
        if isinstance(self._dim, int):
            return "DimensionObject({})".format(self._dim)
        if isinstance(self._dim, DimensionObject):
            return repr(self._dim)
        if isinstance(self._dim, ShapeOperator):
            return "DimensionObject({})".format(repr(self._dim))
        return "DimensionObject('{}')".format(self._dim)

    @staticmethod
    def _same_(obj):
        """
        Returns *obj* if *obj* is @see cl DimensionObject
        otherwise converts it.
        """
        if isinstance(obj, DimensionObject):
            return obj
        return DimensionObject(obj)

    def to_string(self, use_x=True):
        """
        Represents the dimension as a string.
        """
        if isinstance(self._dim, int):
            return '{}'.format(self._dim)
        if isinstance(self._dim, ShapeOperator):
            return self._dim.to_string()
        if isinstance(self._dim, str):
            return self._dim
        if self._dim is None:
            return 'x' if use_x else '?'
        raise NotImplementedError(  # pragma: no cover
            "Not implemented for '{}'.".format(repr(self)))

    def evaluate(self, **kwargs):
        """
        Evalutes the dimension.

        @param  kwargs      value for the variables.
        @return             string or integer
        """
        if isinstance(self._dim, (int, ShapeOperator, DimensionObject)):
            res = self._dim
        elif isinstance(self._dim, str):
            if self._dim in kwargs:
                res = kwargs[self._dim]
            else:
                res = self._dim
        elif self._dim is None:
            pref = str(hex(id(self)))[2:]
            res = "n{}".format(pref)
        elif isinstance(self._dim, ):
            res = self._dim.evaluate(**kwargs)
        else:
            raise NotImplementedError(  # pragma: no cover
                "Not implemented for '{}'.".format(repr(self)))
        if isinstance(res, (ShapeOperator, DimensionObject)):
            return res.evaluate(**kwargs)
        return res

    def __eq__(self, v):
        """
        usual
        """
        if isinstance(v, (int, str)):
            return self._dim == v
        if isinstance(v, DimensionObject):
            return v == self._dim
        if isinstance(v, ShapeOperator):
            ve = v.evaluate()
            return ve == self._dim
        if v is None:
            return self._dim is None
        raise TypeError(  # pragma: no cover
            "Unable to compare a DimensionObject to {}".format(type(v)))

    def __add__(self, obj):
        """
        usual
        """
        return DimensionObject(
            ShapeOperatorAdd(self, DimensionObject._same_(obj)))

    def __mul__(self, obj):
        """
        usual
        """
        return DimensionObject(
            ShapeOperatorMul(self, DimensionObject._same_(obj)))

    def __gt__(self, obj):
        """
        usual
        """
        if obj is None:
            return not isinstance(self._dim, int)
        if isinstance(self._dim, int) and isinstance(obj._dim, int):
            return self._dim > obj._dim
        return DimensionObject(
            ShapeOperatorGreater(self, DimensionObject._same_(obj)))


class ShapeObject(BaseDimensionShape):
    """
    Handles mathematical operations around shapes.
    It stores a type (:epkg:`numpy` type),
    and a name to somehow have an idea of where
    the shape comes from in the :epkg:`ONNX` graph.
    The shape itself is defined by a list of
    @see cl DimensionObject or @see cl ShapeOperator
    or *None* if the shape is unknown. A dimension is an
    integer or a variable encoded as a string. This variable
    is a way to tell the dimension may vary.

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.onnxrt.shape_object import ShapeObject

        sh1 = ShapeObject((1, 2), dtype=numpy.float32)
        sh2 = ShapeObject((45, 2), dtype=numpy.float32)
        mx = max(sh1, sh2)
        print(mx)

        sh1 = ShapeObject((1, 2), dtype=numpy.float32)
        sh2 = ShapeObject((None, 2), dtype=numpy.float32)
        print(sh2)
        mx = max(sh1, sh2)
        print(mx.to_string())

        sh1 = ShapeObject((1, 2), dtype=numpy.float32)
        sh2 = ShapeObject(('n', 2), dtype=numpy.float32)
        print(sh2)
        mx = max(sh1, sh2)
        print(mx.evaluate(n=4))
    """

    def __init__(self, shape, dtype=None, use_n1=False, name=None):
        """
        @param      shape       tuple or `numpy.array`
        @param      dtype       dtype
        @param      use_n1      use `'n'` if the first dimension is unknown
        @param      name        optional, for debugging purposes
        """
        self.name = name
        if isinstance(shape, numpy.ndarray):
            self._shape = [DimensionObject(s) for s in shape.shape]
            self._dtype = shape.dtype
        elif isinstance(shape, dict) and 'type' in shape:
            tshape = shape['type']
            if tshape['kind'] == 'tensor':
                if tshape['shape'] == ('?', ):
                    self._shape = None
                else:
                    self._shape = [DimensionObject(s) for s in tshape['shape']]
                self._dtype = tshape['elem']
            elif tshape['kind'] == 'map':
                self._shape = []
                self._dtype = 'map'
            else:
                raise ValueError(  # pragma: no cover
                    "Wrong shape value {}".format(shape))
        elif isinstance(shape, (tuple, list)):
            self._shape = []
            for s in shape:
                self._shape.append(DimensionObject(s))
            self._dtype = dtype
        elif shape is None:
            # shape is unknown
            self._shape = None
            self._dtype = dtype
        else:
            raise TypeError(  # pragma: no cover
                "Unexpected type for shape: {}".format(type(shape)))

        if self._dtype is None:
            raise ValueError(
                "dtype cannot be None, shape type is {}\n{}".format(
                    type(shape), shape))
        if self._dtype in (float, 'double'):
            self._dtype = numpy.float64
        elif self._dtype in ('float32', 'float'):
            self._dtype = numpy.float32
        elif self._dtype in (numpy.float16, 'float16'):
            self._dtype = numpy.float16
        elif self._dtype in ('int32', ):
            self._dtype = numpy.int32
        elif self._dtype in (int, 'int', 'int64'):
            self._dtype = numpy.int64
        elif self._dtype in (str, 'str'):
            self._dtype = numpy.str
        elif (hasattr(self._dtype, 'type') and self._dtype.type is numpy.string_):
            pass
        elif self._dtype in (bool, 'bool'):
            self._dtype = numpy.bool
        elif self._dtype in (object, numpy.object_):
            pass
        elif self._dtype in (numpy.int8, 'int8', ):
            self._dtype = numpy.int8
        elif self._dtype in (numpy.uint8, 'uint8', ):
            self._dtype = numpy.uint8
        elif self._dtype not in {
                numpy.float32, numpy.float64, numpy.int32, numpy.int64,
                numpy.str, numpy.bool, numpy.float16, None,
                'map'}:
            raise ValueError(  # pragma: no cover
                "dtype has an unexpected value: '{}'.".format(self._dtype))
        if self._shape is not None:
            for i, a in enumerate(self._shape):
                if not isinstance(a, DimensionObject):
                    raise TypeError(  # pragma: no cover
                        'Dimension {} has a wrong type {}'.format(
                            i, type(a)))
            if use_n1:
                sh = self._shape[0] if self._shape else None
                if isinstance(sh, DimensionObject) and sh._dim is None:
                    sh._dim = 'n'

    def reshape(self, shape):
        """
        Creates a new shape, checks the number of elements is the same.
        """
        sh = ShapeObject(shape, self.dtype, getattr(self, '_dim', None),
                         self.name)
        p1 = self.product().evaluate()
        p2 = sh.product().evaluate()
        if isinstance(p1, int) and p1 != p2:
            raise ValueError("Shape {} cannot be reshaped into {} "
                             "(p1={}, p2={}).".format(sh, shape, p1, p2))
        return sh

    def copy(self, dtype=None, name=None):
        """
        A copy not a deepcopy.

        @param      dtype   None or a value to rewrite the type.
        @param      name    overwrites the name
        @return             @see cl ShapeObject
        """
        if self._shape is None:
            return ShapeObject(None, dtype=self.dtype, name=name or self.name)
        return ShapeObject(self._shape.copy(),
                           self.dtype if dtype is None else dtype,
                           name=name or self.name)

    def __getitem__(self, index):
        """
        Extracts a specific dimension.
        """
        if self._shape is None:
            return None
        if isinstance(index, int) and index >= len(self._shape):
            return 1
        return self._shape[index]

    def __setitem__(self, index, value):
        """
        Changes a specific dimension.
        """
        if self._shape is None:
            return
        while len(self._shape) <= index:
            self._shape.append(DimensionObject(1))
        self._shape[index] = value

    @property
    def shape(self):
        """
        Returns the stored shape.
        """
        if self._shape is None:
            return None
        return tuple(self._shape)

    def __len__(self):
        """
        Returns the number of dimensions.
        """
        if self._shape is None:
            return 0
        return len(self._shape)

    @property
    def dtype(self):
        """
        Returns the stored *dtype*.
        """
        return self._dtype

    def reduce(self, axis=1, keepdims=False, dtype=None):
        """
        Reduces the matrix. Removes one dimension.

        @param      axis        axis
        @param      keepdims    keep dimensions, replaces the removed
                                dimension by 1
        @param      dtype       if not None, changes the type
        @return                 new dimension
        """
        if self._shape is None:
            if self.name is None:
                return self.copy()
            return self.copy(name="{}-RD".format(self.name))
        if axis is None:
            return ShapeObject((1, ), self._dtype if dtype is None else dtype,
                               name="{}-RDN".format(self.name))

        if isinstance(axis, ShapeObject):

            def drop_axis(shape, a):
                c = list(shape)
                del c[a[0]]
                return c

            return ShapeObjectFct(
                drop_axis, self, axis, name="DropAxis", dtype=self.dtype)

        if 0 <= axis < len(self._shape):
            cp = self._shape.copy()
            if keepdims:
                cp[axis] = DimensionObject(1)
            else:
                del cp[axis]
            return ShapeObject(cp, self._dtype if dtype is None else dtype,
                               name="{}-RD".format(self.name))
        raise IndexError("axis={} is wrong, shape is {}-tuple and equal to "
                         "{}".format(axis, len(self._shape), self))

    def __repr__(self):
        """
        usual
        """
        st = str(self.dtype)
        if "'" in st:
            st = st.split("'")[1]

        if self.shape is None:
            if self.name is None:
                return "ShapeObject(None, dtype={})".format(st)
            return "ShapeObject(None, dtype={}, name='{}')".format(st, self.name)

        st_shape = []
        for s in self.shape:
            if isinstance(getattr(s, "_dim", None), (int, str)):
                st_shape.append(str(s._dim))
            else:
                st_shape.append(repr(s))
        if len(st_shape) == 1:
            st_shape.append('')
        st_shape = '({})'.format(", ".join(st_shape))
        if self.name is None:
            return "ShapeObject({}, dtype={})".format(st_shape, st)
        return "ShapeObject({}, dtype={}, name='{}')".format(
            st_shape, st, self.name)

    def __iter__(self):
        """
        Iterators over dimensions.
        """
        if self._shape is not None:
            for d in self._shape:
                yield d

    def __gt__(self, a):
        """
        Compares shapes. Operator ``>``.
        """
        if isinstance(a, tuple):
            a = ShapeObject(a, dtype=self._dtype)
        if self._shape is None and a._shape is None:
            return False
        if self._shape is None:
            return True
        if a._shape is None:
            return False
        if len(self) > len(a):
            return True
        if len(self) < len(a):
            return False
        for d1, d2 in zip(self, a):
            if d1 > d2:
                return True
            if d1 < d2:
                return False
        return False

    def __eq__(self, a):
        """
        Tests equality between two shapes.
        """
        if isinstance(a, tuple):
            a = ShapeObject(a, dtype=self._dtype)
        if self._shape is None and a._shape is None:
            return True
        if self._shape is None or a._shape is None:
            return False
        if len(self) != len(a):
            return False
        for d1, d2 in zip(self, a):
            if d1 == d2:
                continue
            return False
        return True

    def evaluate(self, **kwargs):
        """
        Evaluates the shape.
        """
        vs = []
        for v in self:
            d = v.evaluate(**kwargs)
            vs.append(d)
        return ShapeObject(tuple(vs), self._dtype, name="{}-EV".format(self.name))

    def to_string(self, use_x=False):
        """
        Converts shapes into a string.
        """
        shapes = []
        for a in self._shape:
            shapes.append(a.to_string(use_x=use_x))
        return '({})'.format(', '.join(shapes))

    def product(self):
        """
        Multiplies all the dimension.

        @return     @see cl DimensionObject
        """
        cl = self[0]
        for i in range(1, len(self)):
            cl = cl * self[i]
        return cl

    def append(self, dim):
        """
        Appends a dimension.
        """
        if self._shape is None:
            return
        if isinstance(dim, DimensionObject):
            self._shape.append(dim)
        else:
            self._shape.append(DimensionObject(dim))

    def insert(self, dim, pos=0):
        """
        Inserts a dimension at position *pos*.
        """
        if self._shape is None:
            return
        if isinstance(dim, DimensionObject):
            self._shape.insert(pos, dim)
        else:
            self._shape.insert(pos, DimensionObject(dim))

    def squeeze(self, axis):
        """
        Removes one dimension.
        """
        cp = self.copy(name='{}-SZ'.format(self.name))
        cp.drop_axis(axis)
        return cp

    def unsqueeze(self, axes):
        """
        Adds dimensions.
        """
        cp = self
        name = '{}-USZ'.format(self.name)
        for ax in axes[::-1]:
            cp = cp.copy(name=name)
            cp.insert(ax, 1)
        return cp

    def transpose(self, perm):
        """
        Removes one dimension.
        """
        if self.shape is None:
            return self.copy(name='{}-TR'.format(self.name))
        cp = ShapeObject([None for p in perm], dtype=self.dtype,
                         name="{}-TR".format(self.name))
        for i, p in enumerate(perm):
            if p >= len(self):
                # This should not happen.
                cp._shape[i] = None
            else:
                cp._shape[i] = self._shape[p]
        return cp

    def drop_axis(self, axis):
        """
        Drops an axis.
        """
        if self._shape is not None:
            if isinstance(axis, (tuple, list)):
                for i in sorted(axis, reverse=True):
                    del self._shape[i]
            else:
                del self._shape[axis]

    def broadcast(self, a):
        """
        Computes the shape after a broadcast.
        """
        if a is None:
            raise ValueError("a should not be None")  # pragma: no cover
        if a._shape is None:
            return a.copy()
        if self._shape is None:
            return self.copy()
        mx = max(len(self._shape), len(a._shape))
        res = []
        for i in range(mx):
            if i < len(self._shape):
                if i < len(a._shape):
                    res.append(ShapeOperatorMax(self[i], a[i]))
                else:
                    res.append(self[i])
            else:
                res.append(a[i])
        return ShapeObject(tuple(res), self.dtype, False,
                           name="broadcast-{}-{}".format(self.name, a.name))

    @staticmethod
    def _infer_merged_type(*args):
        tys = set(a.dtype for a in args)
        if len(tys) == 1:
            return list(tys)[0]
        if any(tys & {numpy.float64, numpy.int64,
                      numpy.float32, numpy.int32,
                      numpy.float16}):
            return numpy.float64
        raise RuntimeError(  # pragma: no cover
            "Unable to infer types based on {}.".format(tys))

    def concat_columns(self, axis, *shapes):
        """
        Concatenates columns from *shapes* to this one
        along one axis.
        """
        args = [self] + list(shapes)
        dtype = self._infer_merged_type(*args)
        dim_axis = args[0][axis]
        if dim_axis is None:
            return ShapeObject(None, dtype=dtype)
        for a in shapes:
            if a[axis] is None:
                return ShapeObject(None, dtype=dtype)
            dim_axis = dim_axis + a[axis]
        a0 = args[0].copy(dtype=dtype)
        a0[axis] = dim_axis
        return a0

    @staticmethod
    def einsum_shape(equation, *inputs):
        """
        Computes :epkg:`einsum` shapes.
        Not the most efficient one as it creates variables
        of the given shapes.
        """
        for inp in inputs:
            if inp.shape is None:
                return inp
        inp, out = [_.strip() for _ in equation.split(b"->")]
        inps = [_.strip() for _ in inp.split(b',')]
        if len(inputs) != len(inps):
            raise RuntimeError(  # pragma: no cover
                "Input mismatch between '{}' and {}.".format(equation, inps))
        shs = {}
        for a, b in zip(inps, inputs):
            if len(a) != len(b):
                raise RuntimeError(  # pragma: no cover
                    "Input mismatch '{}' (in '{}') and {}.".format(a, equation, b))
            for c, s in zip(a, b):
                if c not in shs:
                    shs[c] = s
                elif shs[c] != s:
                    raise RuntimeError(  # pragma: no cover
                        "Equation '{}'. Dimension mismatch '{}' != {}.".format(
                            equation, s, shs[c]))
        new_shape = [shs[i] for i in out]
        return ShapeObject(new_shape, dtype=ShapeObject._infer_merged_type(*inputs))

    @staticmethod
    def gather_shape(input, indices, axis):
        """
        Computes Gather shapes.
        """
        input_rank = len(input)
        if input_rank is None:
            return ShapeObject(None, dtype=input._dtype)
        index_rank = len(indices)
        if index_rank is None:
            return ShapeObject(None, dtype=input._dtype)

        if axis < 0:
            axis = input_rank + axis

        shape = []
        for i in range(axis):
            shape.append(input[i])

        for dim in indices:
            shape.append(dim)

        for i in range(axis + 1, input_rank):
            shape.append(input[i])

        return ShapeObject(shape, dtype=input._dtype)


class ShapeObjectFct(ShapeObject):
    """
    Computes a shape depending on a user defined function.
    See @see cl Conv for an example.
    """

    def __init__(self, fct, *shapes, dtype=None, name=None):
        """
        @param      fct         function
        @param      shapes      shapes sent to fct
        @param      dtype       dtype
        @param      name        optional, for debugging purposes
        """
        ShapeObject.__init__(self, None, dtype=dtype, name=name)
        self._fct = fct
        self._shapes = shapes

    def evaluate(self, **kwargs):
        """
        Evaluates the shape.
        """
        vs = []
        for v in self._shapes:
            d = v.evaluate(**kwargs)
            vs.append(d)
        res = self._fct(*vs)
        if self.name is not None:
            res.name = self.name
        return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRun


class Mean(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, *args):  # pylint: disable=W0221
        if self.inplaces.get(0, False):
            return self._run_inplace(*args)
        res = args[0].copy()
        for m in args[1:]:
            res += m
        return (res / len(args), )

    def _run_inplace(self, *args):
        res = args[0]
        for m in args[1:]:
            res += m
        return (res / len(args), )

    def _infer_shapes(self, *args):  # pylint: disable=W0221
        return (args[0], )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import DimensionObject


class OneHotEncoder(OpRun):
    """
    :epkg:`ONNX` specifications does not mention
    the possibility to change the output type,
    sparse, dense, float, double.
    """

    atts = {'cats_int64s': numpy.empty(0, dtype=numpy.int64),
            'cats_strings': numpy.empty(0, dtype=numpy.str),
            'zeros': 1,
            }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=OneHotEncoder.atts,
                       **options)
        if len(self.cats_int64s) > 0:
            self.classes_ = {v: i for i, v in enumerate(self.cats_int64s)}
        elif len(self.cats_strings) > 0:
            self.classes_ = {v.decode('utf-8'): i for i,
                             v in enumerate(self.cats_strings)}
        else:
            raise RuntimeError("No encoding was defined.")  # pragma: no cover

    def _run(self, x):  # pylint: disable=W0221
        shape = x.shape
        new_shape = shape + (len(self.classes_), )
        res = numpy.zeros(new_shape, dtype=numpy.float32)
        if len(x.shape) == 1:
            for i, v in enumerate(x):
                j = self.classes_.get(v, -1)
                if j >= 0:
                    res[i, j] = 1.
        elif len(x.shape) == 2:
            for a, row in enumerate(x):
                for i, v in enumerate(row):
                    j = self.classes_.get(v, -1)
                    if j >= 0:
                        res[a, i, j] = 1.
        else:
            raise RuntimeError(  # pragma: no cover
                "This operator is not implemented for shape {}.".format(x.shape))

        if not self.zeros:
            red = res.sum(axis=len(res.shape) - 1)
            if numpy.min(red) == 0:
                rows = []
                for i, val in enumerate(red):
                    if val == 0:
                        rows.append(dict(row=i, value=x[i]))
                        if len(rows) > 5:
                            break
                raise RuntimeError(  # pragma no cover
                    "One observation did not have any defined category.\n"
                    "classes: {}\nfirst rows:\n{}\nres:\n{}\nx:\n{}".format(
                        self.classes_, "\n".join(str(_) for _ in rows),
                        res[:5], x[:5]))

        return (res, )

    def _infer_shapes(self, x):  # pylint: disable=W0221
        new_shape = x.copy()
        dim = DimensionObject(len(self.classes_))
        new_shape.append(dim)
        new_shape._dtype = numpy.float32
        new_shape.name = self.onnx_node.name
        return (new_shape, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject
from .op_gather_ import (  # pylint: disable=E0611,E0401
    GatherFloat, GatherDouble, GatherInt64)


class Gather(OpRun):

    atts = {'axis': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Gather.atts,
                       **options)
        self.rt_ = {
            'float32': GatherFloat(self.axis),
            'float64': GatherDouble(self.axis),
            'int64': GatherInt64(self.axis)}

    def _run(self, x, indices):  # pylint: disable=W0221
        if not x.flags['C_CONTIGUOUS']:
            x = numpy.ascontiguousarray(x)
        if not indices.flags['C_CONTIGUOUS']:
            indices = indices.ascontiguousarray()
        try:
            return (self.rt_[str(x.dtype)].compute(x, indices), )
        except KeyError:
            return (numpy.take(x, indices, axis=self.axis), )

    def _infer_shapes(self, x, indices):  # pylint: disable=E0202,W0221
        """
        Returns the same shape by default.
        """
        return (ShapeObject.gather_shape(x, indices, self.axis), )



# -*- encoding: utf-8 -*-
"""
@file
@brief Overloads a conversion function.
"""
import pprint
from collections import OrderedDict
import numpy
import pandas
try:
    from sklearn.metrics._scorer import _PredictScorer
except ImportError:  # pragma: no cover
    # scikit-learn < 0.22
    from sklearn.metrics.scorer import _PredictScorer
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from skl2onnx.common.data_types import (
    FloatTensorType, DoubleTensorType, DataType, guess_numpy_type,
    StringTensorType, Int64TensorType)
from skl2onnx import convert_sklearn
from skl2onnx.algebra.onnx_operator_mixin import OnnxOperatorMixin
from skl2onnx.algebra.type_helper import _guess_type
from .register_rewritten_converters import register_rewritten_operators
from .register import register_converters
from .scorers import CustomScorerTransform


def convert_scorer(fct, initial_types, name=None,
                   target_opset=None, options=None,
                   custom_conversion_functions=None,
                   custom_shape_calculators=None,
                   custom_parsers=None, white_op=None,
                   black_op=None, final_types=None):
    """
    Converts a scorer into :epkg:`ONNX` assuming
    there exists a converter associated to it.
    The function wraps the function into a custom
    transformer, then calls function *convert_sklearn*
    from :epkg:`sklearn-onnx`.

    @param  fct                         function to convert (or a scorer from
                                        :epkg:`scikit-learn`)
    @param  initial_types               types information
    @param  name                        name of the produced model
    @param  target_opset                to do it with a different target opset
    @param  options                     additional parameters for the conversion
    @param  custom_conversion_functions a dictionary for specifying the user customized
                                        conversion function, it takes precedence over
                                        registered converters
    @param  custom_shape_calculators    a dictionary for specifying the user
                                        customized shape calculator
                                        it takes precedence over registered
                                        shape calculators.
    @param  custom_parsers              parsers determine which outputs is expected
                                        for which particular task, default parsers are
                                        defined for classifiers, regressors, pipeline but
                                        they can be rewritten, *custom_parsers* is a dictionary
                                        ``{ type: fct_parser(scope, model, inputs,
                                        custom_parsers=None) }``
    @param      white_op                white list of ONNX nodes allowed
                                        while converting a pipeline, if empty,
                                        all are allowed
    @param      black_op                black list of ONNX nodes allowed
                                        while converting a pipeline, if empty,
                                        none are blacklisted
    @param      final_types             a python list. Works the same way as
                                        initial_types but not mandatory, it is used
                                        to overwrites the type (if type is not None)
                                        and the name of every output.
    @return                             :epkg:`ONNX` graph
    """
    if hasattr(fct, '_score_func'):
        kwargs = fct._kwargs
        fct = fct._score_func
    else:
        kwargs = None  # pragma: no cover
    if name is None:
        name = "mlprodict_fct_ONNX(%s)" % fct.__name__
    tr = CustomScorerTransform(fct.__name__, fct, kwargs)
    return convert_sklearn(
        tr, initial_types=initial_types,
        target_opset=target_opset, options=options,
        custom_conversion_functions=custom_conversion_functions,
        custom_shape_calculators=custom_shape_calculators,
        custom_parsers=custom_parsers, white_op=white_op,
        black_op=black_op, final_types=final_types)


def guess_initial_types(X, initial_types):
    """
    Guesses initial types from an array or a dataframe.

    @param      X               array or dataframe
    @param      initial_types   hints about X
    @return                     data types
    """
    if X is None and initial_types is None:
        raise NotImplementedError(  # pragma: no cover
            "Initial types must be specified.")
    elif initial_types is None:
        if isinstance(X, (numpy.ndarray, pandas.DataFrame)):
            X = X[:1]
        if isinstance(X, pandas.DataFrame):
            initial_types = []
            for c in X.columns:
                if isinstance(X[c].values[0], (str, numpy.str)):
                    g = StringTensorType()
                else:
                    g = _guess_type(X[c].values)
                g.shape = [None, 1]
                initial_types.append((c, g))
        else:
            gt = _guess_type(X)
            initial_types = [('X', gt)]
    return initial_types


def _replace_tensor_type(schema, tensor_type):
    res = []
    for name, ty in schema:
        cl = ty.__class__
        if cl in (FloatTensorType, DoubleTensorType) and cl != tensor_type:
            ty = tensor_type(ty.shape)
        res.append((name, ty))
    return res


def guess_schema_from_data(X, tensor_type=None, schema=None):
    """
    Guesses initial types from a dataset.

    @param      X               dataset (dataframe, array)
    @param      tensor_type     if not None, replaces every
                                *FloatTensorType* or *DoubleTensorType*
                                by this one
    @param      schema          known schema
    @return                     schema (list of typed and named columns)
    """
    init = guess_initial_types(X, schema)
    if tensor_type is not None:
        init = _replace_tensor_type(init, tensor_type)
    # Grouping column
    unique = set()
    for _, col in init:
        if len(col.shape) != 2:
            return init  # pragma: no cover
        if col.shape[0] is not None:
            return init  # pragma: no cover
        if len(unique) > 0 and col.__class__ not in unique:
            return init  # pragma: no cover
        unique.add(col.__class__)
    unique = list(unique)
    return [('X', unique[0]([None, sum(_[1].shape[1] for _ in init)]))]


def get_inputs_from_data(X, schema=None):
    """
    Produces input data for *onnx* runtime.

    @param  X       data
    @param  schema  schema if None, schema is guessed with
        @see fn guess_schema_from_data
    @return input data
    """
    def _cast_data(X, ct):
        if isinstance(ct, FloatTensorType):
            return X.astype(numpy.float32)
        if isinstance(ct, DoubleTensorType):
            return X.astype(numpy.float64)
        if isinstance(ct, StringTensorType):
            return X.astype(numpy.str)
        if isinstance(ct, Int64TensorType):
            return X.astype(numpy.int64)
        raise RuntimeError(
            "Unexpected column type {} for type {}."
            "".format(ct, type(X)))

    if schema is None:
        schema = guess_schema_from_data(X)
    if isinstance(X, numpy.ndarray):
        if len(schema) != 1:
            raise RuntimeError(  # pragma: no cover
                "More than one column but input is an array.")
        return {schema[0][0]: _cast_data(X, schema[0][1])}
    elif isinstance(X, pandas.DataFrame):
        if len(schema) != X.shape[1]:
            raise RuntimeError(  # pragma: no cover
                "Mismatch between onnx columns {} and DataFrame columns {}"
                "".format(len(schema), X.shape[1]))
        return {sch[0]: _cast_data(X[c].values, sch[1]).reshape((-1, 1))
                for sch, c in zip(schema, X.columns)}
    else:
        raise TypeError(
            "Unexpected type {}, expecting an array or a dataframe."
            "".format(type(X)))


def guess_schema_from_model(model, tensor_type=None, schema=None):
    """
    Guesses initial types from a model.

    @param      X               dataset (dataframe, array)
    @param      tensor_type     if not None, replaces every
                                *FloatTensorType* or *DoubleTensorType*
                                by this one
    @param      schema          known schema
    @return                     schema (list of typed and named columns)
    """
    if schema is not None:
        try:
            guessed = guess_schema_from_model(model)
        except NotImplementedError:  # pragma: no cover
            return _replace_tensor_type(schema, tensor_type)
        if len(guessed) != len(schema):
            raise RuntimeError(  # pragma: no cover
                "Given schema and guessed schema are not the same:\nGOT: {}\n-----\nGOT:\n{}".format(
                    schema, guessed))
        return _replace_tensor_type(schema, tensor_type)

    if hasattr(model, 'coef_'):
        # linear model
        init = [('X', FloatTensorType([None, model.coef_.shape[1]]))]
        return _replace_tensor_type(init, tensor_type)
    elif hasattr(model, 'dump_model'):
        dumped = model.dump_model()
        if isinstance(dumped, dict) and 'feature_names' in dumped:
            names = dumped['feature_names']
            init = [(name, FloatTensorType([None, 1])) for name in names]
            return _replace_tensor_type(init, tensor_type)

    data = pprint.pformat(model.__dict__)
    dirs = pprint.pformat(dir(model))
    if hasattr(model, 'dump_model'):  # pragma: no cover
        dumped = model.dump_model()
        keys = list(sorted(dumped))
        last = pprint.pformat([keys, dumped])
        if len(last) >= 200000:
            last = last[:200000] + "\n..."
    else:
        last = ""
    raise NotImplementedError(  # pragma: no cover
        "Unable to guess schema for model {}\n{}\n----\n{}\n------\n{}".format(
            model.__class__, data, dirs, last))


def to_onnx(model, X=None, name=None, initial_types=None,
            target_opset=None, options=None, rewrite_ops=False,
            white_op=None, black_op=None, final_types=None):
    """
    Converts a model using on :epkg:`sklearn-onnx`.

    @param      model           model to convert or a function
                                wrapped into :epkg:`_PredictScorer` with
                                function :epkg:`make_scorer`
    @param      X               training set (at least one row),
                                can be None, it is used to infered the
                                input types (*initial_types*)
    @param      initial_types   if *X* is None, then *initial_types* must be
                                defined
    @param      name            name of the produced model
    @param      target_opset    to do it with a different target opset
    @param      options         additional parameters for the conversion
    @param      rewrite_ops     rewrites some existing converters,
                                the changes are permanent
    @param      white_op        white list of ONNX nodes allowed
                                while converting a pipeline, if empty,
                                all are allowed
    @param      black_op        black list of ONNX nodes allowed
                                while converting a pipeline, if empty,
                                none are blacklisted
    @param      final_types     a python list. Works the same way as
                                initial_types but not mandatory, it is used
                                to overwrites the type (if type is not None)
                                and the name of every output.
    @return                     converted model

    The function rewrites function *to_onnx* from :epkg:`sklearn-onnx`
    but may changes a few converters if *rewrite_ops* is True.
    For example, :epkg:`ONNX` only supports *TreeEnsembleRegressor*
    for float but not for double. It becomes available
    if ``rewrite_ops=True``.

    .. faqref::
        :title: How to deal with a dataframe as input?

        Each column of the dataframe is considered as an named input.
        The first step is to make sure that every column type is correct.
        :epkg:`pandas` tends to select the least generic type to
        hold the content of one column. :epkg:`ONNX` does not automatically
        cast the data it receives. The data must have the same type with
        the model is converted and when the converted model receives
        the data to predict.

        .. runpython::
            :showcode:

            from io import StringIO
            from textwrap import dedent
            import numpy
            import pandas
            from pyquickhelper.pycode import ExtTestCase
            from sklearn.preprocessing import OneHotEncoder
            from sklearn.pipeline import Pipeline
            from sklearn.compose import ColumnTransformer
            from mlprodict.onnx_conv import to_onnx
            from mlprodict.onnxrt import OnnxInference

            text = dedent('''
                __SCHEMA__
                7.4,0.7,0.0,1.9,0.076,11.0,34.0,0.9978,3.51,0.56,9.4,5,red
                7.8,0.88,0.0,2.6,0.098,25.0,67.0,0.9968,3.2,0.68,9.8,5,red
                7.8,0.76,0.04,2.3,0.092,15.0,54.0,0.997,3.26,0.65,9.8,5,red
                11.2,0.28,0.56,1.9,0.075,17.0,60.0,0.998,3.16,0.58,9.8,6,red
                ''')
            text = text.replace(
                "__SCHEMA__",
                "fixed_acidity,volatile_acidity,citric_acid,residual_sugar,chlorides,"
                "free_sulfur_dioxide,total_sulfur_dioxide,density,pH,sulphates,"
                "alcohol,quality,color")

            X_train = pandas.read_csv(StringIO(text))
            for c in X_train.columns:
                if c != 'color':
                    X_train[c] = X_train[c].astype(numpy.float32)
            numeric_features = [c for c in X_train if c != 'color']

            pipe = Pipeline([
                ("prep", ColumnTransformer([
                    ("color", Pipeline([
                        ('one', OneHotEncoder()),
                        ('select', ColumnTransformer(
                            [('sel1', 'passthrough', [0])]))
                    ]), ['color']),
                    ("others", "passthrough", numeric_features)
                ])),
            ])

            pipe.fit(X_train)
            pred = pipe.transform(X_train)
            print(pred)

            model_onnx = to_onnx(pipe, X_train, target_opset=12)
            oinf = OnnxInference(model_onnx)

            # The dataframe is converted into a dictionary,
            # each key is a column name, each value is a numpy array.
            inputs = {c: X_train[c].values for c in X_train.columns}
            inputs = {c: v.reshape((v.shape[0], 1)) for c, v in inputs.items()}

            onxp = oinf.run(inputs)
            print(onxp)
    """
    if isinstance(model, OnnxOperatorMixin):
        if not hasattr(model, 'op_version'):
            raise RuntimeError(  # pragma: no cover
                "Missing attribute 'op_version' for type '{}'.".format(
                    type(model)))
        return model.to_onnx(
            X=X, name=name, options=options, black_op=black_op,
            white_op=white_op, final_types=final_types)
    if rewrite_ops:
        old_values = register_rewritten_operators()
        register_converters()
    else:
        old_values = None

    def _guess_type_(X, itype, dtype):
        initial_types = guess_initial_types(X, itype)
        if dtype is None:
            if hasattr(X, 'dtypes'):  # DataFrame
                dtype = numpy.float32
            elif hasattr(X, 'dtype'):
                dtype = X.dtype
            elif hasattr(X, 'type'):
                dtype = guess_numpy_type(X.type)
            elif initial_types is not None:
                dtype = guess_numpy_type(initial_types[0][1])
            else:
                raise RuntimeError(  # pragma: no cover
                    "dtype cannot be guessed: {}".format(
                        type(X)))
            if dtype != numpy.float64:
                dtype = numpy.float32
        if dtype is None:
            raise RuntimeError("dtype cannot be None")  # pragma: no cover
        if isinstance(dtype, FloatTensorType):
            dtype = numpy.float32  # pragma: no cover
        elif isinstance(dtype, DoubleTensorType):
            dtype = numpy.float64  # pragma: no cover
        new_dtype = dtype
        if isinstance(dtype, numpy.ndarray):
            new_dtype = dtype.dtype  # pragma: no cover
        elif isinstance(dtype, DataType):
            new_dtype = numpy.float32  # pragma: no cover
        if new_dtype not in (numpy.float32, numpy.float64, numpy.int64,
                             numpy.int32, numpy.float16):
            raise NotImplementedError(  # pragma: no cover
                "dtype should be real not {} ({})".format(new_dtype, dtype))
        return initial_types, dtype, new_dtype

    if isinstance(model, _PredictScorer):
        if X is not None and not isinstance(X, OrderedDict):
            raise ValueError("For a scorer, parameter X should be a OrderedDict not {}."
                             "".format(type(X)))
        if initial_types is None:
            dts = []
            initial_types = []
            for k, v in X.items():
                if hasattr(v, 'dtype'):
                    dtype = guess_numpy_type(v.dtype)
                else:
                    dtype = v  # pragma: no cover
                it, _, ndt = _guess_type_(v, None, dtype)
                for i in range(len(it)):  # pylint: disable=C0200
                    it[i] = (k, it[i][1])  # pylint: disable=C0200
                initial_types.extend(it)
                dts.append(ndt)
            ndt = set(dts)
            if len(ndt) != 1:
                raise RuntimeError(  # pragma: no cover
                    "Multiple dtype is not efficient {}.".format(ndt))
        res = convert_scorer(model, initial_types, name=name,
                             target_opset=target_opset, options=options,
                             black_op=black_op, white_op=white_op,
                             final_types=final_types)
    else:
        if name is None:
            name = "mlprodict_ONNX(%s)" % model.__class__.__name__

        initial_types, dtype, _ = _guess_type_(X, initial_types, None)
        res = convert_sklearn(model, initial_types=initial_types, name=name,
                              target_opset=target_opset, options=options,
                              black_op=black_op, white_op=white_op,
                              final_types=final_types)

    if old_values is not None:
        register_rewritten_operators(old_values)
    return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _batchnorm_test_mode(x, s, bias, mean, var, epsilon=1e-5):
    dims_x = len(x.shape)
    dim_ones = (1,) * (dims_x - 2)
    s = s.reshape(-1, *dim_ones)
    bias = bias.reshape(-1, *dim_ones)
    mean = mean.reshape(-1, *dim_ones)
    var = var.reshape(-1, *dim_ones)
    return s * (x - mean) / numpy.sqrt(var + epsilon) + bias


class BatchNormalization(OpRun):

    atts = {'epsilon': 1e-5, 'momentum': 0.9}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=BatchNormalization.atts,
                       **options)

    def _run(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        res = _batchnorm_test_mode(
            x, scale, bias, mean, var, epsilon=self.epsilon)
        return (res, )

    def _infer_shapes(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        return (x, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject


class Einsum(OpRun):

    atts = {'equation': ''}
    python_inputs = ['*inputs']

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Einsum.atts,
                       **options)

    def _run(self, *args):  # pylint: disable=W0221
        return (numpy.einsum(self.equation, *args), )

    def _infer_shapes(self, *args):  # pylint: disable=W0221
        return (ShapeObject.einsum_shape(self.equation, *args), )

    def _infer_type(self, *args):
        return ShapeObject._infer_merged_type(*args)

    def to_python(self, inputs):
        return "import numpy", "return numpy.einsum(equation, *inputs)"



"""
@file
@brief Rewrites some of the converters implemented in
:epkg:`sklearn-onnx`.
"""
from skl2onnx.common._registration import _converter_pool

try:
    from skl2onnx.common._registration import RegisteredConverter
except ImportError:  # pragma: no cover
    # sklearn-onnx <= 1.6.0
    RegisteredConverter = lambda fct, opts: fct

from .sklconv.tree_converters import (
    new_convert_sklearn_decision_tree_classifier,
    new_convert_sklearn_decision_tree_regressor,
    new_convert_sklearn_gradient_boosting_classifier,
    new_convert_sklearn_gradient_boosting_regressor,
    new_convert_sklearn_random_forest_classifier,
    new_convert_sklearn_random_forest_regressor,
)
from .sklconv.svm_converters import (
    new_convert_sklearn_svm_classifier,
    new_convert_sklearn_svm_regressor,
)


_overwritten_operators = {
    #
    'SklearnOneClassSVM': RegisteredConverter(
        new_convert_sklearn_svm_regressor,
        _converter_pool['SklearnOneClassSVM'].get_allowed_options()),
    'SklearnSVR': RegisteredConverter(
        new_convert_sklearn_svm_regressor,
        _converter_pool['SklearnSVR'].get_allowed_options()),
    'SklearnSVC': RegisteredConverter(
        new_convert_sklearn_svm_classifier,
        _converter_pool['SklearnSVC'].get_allowed_options()),
    #
    'SklearnDecisionTreeRegressor': RegisteredConverter(
        new_convert_sklearn_decision_tree_regressor,
        _converter_pool['SklearnDecisionTreeRegressor'].get_allowed_options()),
    'SklearnDecisionTreeClassifier': RegisteredConverter(
        new_convert_sklearn_decision_tree_classifier,
        _converter_pool['SklearnDecisionTreeClassifier'].get_allowed_options()),
    #
    'SklearnExtraTreeRegressor': RegisteredConverter(
        new_convert_sklearn_decision_tree_regressor,
        _converter_pool['SklearnExtraTreeRegressor'].get_allowed_options()),
    'SklearnExtraTreeClassifier': RegisteredConverter(
        new_convert_sklearn_decision_tree_classifier,
        _converter_pool['SklearnExtraTreeClassifier'].get_allowed_options()),
    #
    'SklearnExtraTreesRegressor': RegisteredConverter(
        new_convert_sklearn_random_forest_regressor,
        _converter_pool['SklearnExtraTreesRegressor'].get_allowed_options()),
    'SklearnExtraTreesClassifier': RegisteredConverter(
        new_convert_sklearn_random_forest_classifier,
        _converter_pool['SklearnExtraTreesClassifier'].get_allowed_options()),
    #
    'SklearnGradientBoostingRegressor': RegisteredConverter(
        new_convert_sklearn_gradient_boosting_regressor,
        _converter_pool['SklearnGradientBoostingRegressor'].get_allowed_options()),
    'SklearnGradientBoostingClassifier': RegisteredConverter(
        new_convert_sklearn_gradient_boosting_classifier,
        _converter_pool['SklearnGradientBoostingClassifier'].get_allowed_options()),
    #
    'SklearnHistGradientBoostingRegressor': RegisteredConverter(
        new_convert_sklearn_random_forest_regressor,
        _converter_pool['SklearnHistGradientBoostingRegressor'].get_allowed_options()),
    'SklearnHistGradientBoostingClassifier': RegisteredConverter(
        new_convert_sklearn_random_forest_classifier,
        _converter_pool['SklearnHistGradientBoostingClassifier'].get_allowed_options()),
    #
    'SklearnRandomForestRegressor': RegisteredConverter(
        new_convert_sklearn_random_forest_regressor,
        _converter_pool['SklearnRandomForestRegressor'].get_allowed_options()),
    'SklearnRandomForestClassifier': RegisteredConverter(
        new_convert_sklearn_random_forest_classifier,
        _converter_pool['SklearnRandomForestClassifier'].get_allowed_options()),
}


def register_rewritten_operators(new_values=None):
    """
    Registers modified operators and returns the old values.

    @param      new_values      operators to rewrite or None
                                to rewrite default ones
    @return                     old values
    """
    if new_values is None:
        for rew in _overwritten_operators:
            if rew not in _converter_pool:
                raise KeyError(  # pragma: no cover
                    "skl2onnx was not imported and '{}' was not registered."
                    "".format(rew))
        old_values = {k: _converter_pool[k] for k in _overwritten_operators}
        _converter_pool.update(_overwritten_operators)
        return old_values

    for rew in new_values:
        if rew not in _converter_pool:
            raise KeyError(  # pragma: no cover
                "skl2onnx was not imported and '{}' was not registered."
                "".format(rew))
    old_values = {k: _converter_pool[k] for k in new_values}
    _converter_pool.update(new_values)
    return old_values
